"""Test auth functions."""

import json
import logging
from os import listdir
from os.path import isfile, join
from unittest.mock import patch

import aiohttp
import jwt
import pytest
from aioresponses import aioresponses

from pyenphase import Envoy, EnvoyTokenAuth
from pyenphase.auth import EnvoyLegacyAuth
from pyenphase.const import (
    URL_AUTH_CHECK_JWT,
    URL_DEVICE_DATA,
    URL_PRODUCTION_INVERTERS,
    SupportedFeatures,
)
from pyenphase.exceptions import EnvoyAuthenticationError, EnvoyAuthenticationRequired

from .common import (
    get_mock_envoy,
    load_fixture,
    load_json_fixture,
    prep_envoy,
    start_7_firmware_mock,
    temporary_log_level,
)

LOGGER = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_wrong_auth_order_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test data collected fails before auth is done"""
    version = "7.6.175_standard"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)
    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()

    with pytest.raises(EnvoyAuthenticationRequired):
        await envoy.update()

    # now in correct order
    await envoy.authenticate("username", "password")
    data = await envoy.update()
    assert data


@pytest.mark.asyncio
async def test_with_3_9_36_firmware_bad_auth(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Verify with 3.9.36 firmware with incorrect auth."""
    version = "3.9.36_bad_auth"
    mock_aioresponse.get(
        "https://127.0.0.1/info", status=200, body=await load_fixture(version, "info")
    )
    mock_aioresponse.get("https://127.0.0.1/info.xml", status=200, body="")
    mock_aioresponse.get("https://127.0.0.1/production", status=404)
    mock_aioresponse.get("https://127.0.0.1/production.json", status=404)
    mock_aioresponse.get(
        "https://127.0.0.1/api/v1/production",
        status=401,
        payload=await load_json_fixture(version, "api_v1_production"),
    )
    mock_aioresponse.get(
        "https://127.0.0.1/api/v1/production/inverters",
        status=200,
        payload=await load_json_fixture(version, "api_v1_production_inverters"),
    )
    mock_aioresponse.get(
        "https://127.0.0.1/ivp/ensemble/inventory", status=200, payload=[]
    )

    path = f"tests/fixtures/{version}"
    files = [f for f in listdir(path) if isfile(join(path, f))]
    if "admin_lib_tariff" in files:
        try:
            json_data = await load_json_fixture(version, "admin_lib_tariff")
        except json.decoder.JSONDecodeError:
            json_data = None
        mock_aioresponse.get(
            "https://127.0.0.1/admin/lib/tariff", status=200, payload=json_data
        )
    else:
        mock_aioresponse.get("https://127.0.0.1/admin/lib/tariff", status=401)

    mock_aioresponse.get("https://127.0.0.1/ivp/meters", status=200, payload=[])

    # Add the HTTP version of api/v1/production with 401 as well
    mock_aioresponse.get(
        "http://127.0.0.1/api/v1/production",
        status=401,
        payload=await load_json_fixture(version, "api_v1_production"),
    )

    # Add other required endpoints for the probe
    mock_aioresponse.get("https://127.0.0.1/production.json?details=1", status=404)
    mock_aioresponse.get("http://127.0.0.1/production.json?details=1", status=404)
    mock_aioresponse.get(
        "http://127.0.0.1/production",
        status=200,
        body=await load_fixture(version, "production"),
    )

    with pytest.raises(EnvoyAuthenticationRequired):
        await get_mock_envoy(test_client_session)


@pytest.mark.asyncio
async def test_production_with_3_9_36_firmware_bad_auth(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test Authentication failed for http://127.0.0.1/api/v1/production."""
    version = "3.9.36_bad_auth"
    mock_aioresponse.get(
        "https://127.0.0.1/info", status=200, body=await load_fixture(version, "info")
    )
    mock_aioresponse.get("https://127.0.0.1/info.xml", status=200, body="")
    mock_aioresponse.get("https://127.0.0.1/production", status=404)
    mock_aioresponse.get("https://127.0.0.1/production.json", status=404)
    mock_aioresponse.get(
        "https://127.0.0.1/api/v1/production",
        status=401,
        payload=await load_json_fixture(version, "api_v1_production"),
    )
    mock_aioresponse.get(
        "https://127.0.0.1/api/v1/production/inverters",
        status=200,
        payload=await load_json_fixture(version, "api_v1_production_inverters"),
    )
    mock_aioresponse.get(
        "https://127.0.0.1/ivp/ensemble/inventory", status=200, payload=[]
    )

    path = f"tests/fixtures/{version}"
    files = [f for f in listdir(path) if isfile(join(path, f))]
    if "admin_lib_tariff" in files:
        try:
            json_data = await load_json_fixture(version, "admin_lib_tariff")
        except json.decoder.JSONDecodeError:
            json_data = None
        mock_aioresponse.get(
            "https://127.0.0.1/admin/lib/tariff", status=200, payload=json_data
        )
    else:
        mock_aioresponse.get("https://127.0.0.1/admin/lib/tariff", status=401)

    mock_aioresponse.get("https://127.0.0.1/ivp/meters", status=200, payload=[])

    # Add the HTTP version of api/v1/production with 401 as well
    mock_aioresponse.get(
        "http://127.0.0.1/api/v1/production",
        status=401,
        payload=await load_json_fixture(version, "api_v1_production"),
    )

    # Add other required endpoints for the probe
    mock_aioresponse.get("https://127.0.0.1/production.json?details=1", status=404)
    mock_aioresponse.get("http://127.0.0.1/production.json?details=1", status=404)
    mock_aioresponse.get(
        "http://127.0.0.1/production",
        status=200,
        body=await load_fixture(version, "production"),
    )

    with pytest.raises(EnvoyAuthenticationRequired):
        await get_mock_envoy(test_client_session)


@pytest.mark.parametrize(
    (
        "username",
        "password",
    ),
    [
        ("installer", ""),
        ("envoy", ""),
    ],
)
@pytest.mark.asyncio
async def test_known_users_with_3_9_36_firmware(
    username: str,
    password: str,
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
) -> None:
    """Test successful login with known usernames."""
    version = "3.9.36"
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    await envoy.authenticate(username, password)

    # test cookies function now cookies are not on request
    assert envoy.auth
    used_cookies = envoy.auth.cookies
    assert used_cookies == {}

    data = await envoy.update()
    assert data

    # cov: force test failure of Digest authentication for local Envoy.
    assert isinstance(envoy.auth, EnvoyLegacyAuth)
    envoy.auth.local_password = ""
    assert envoy.auth.auth is None


@pytest.mark.asyncio
async def test_unknown_user_with_3_9_36_firmware(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test Could not setup authentication object with 3.9.x"""
    version = "3.9.36"
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    with pytest.raises(EnvoyAuthenticationRequired):
        await envoy.authenticate("unknown", None)


@pytest.mark.parametrize(
    (
        "username",
        "password",
    ),
    [
        ("installer", ""),
        ("envoy", ""),
        ("unknown", ""),
    ],
)
@pytest.mark.asyncio
async def test_blank_passwords_with_7_6_175_standard(
    username: str,
    password: str,
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
) -> None:
    """Test Could not setup authentication object with 7.6.x"""
    version = "7.6.175_standard"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    envoy = Envoy("127.0.0.1", client=test_client_session)

    await envoy.setup()

    with pytest.raises(EnvoyAuthenticationRequired):
        await envoy.authenticate(username, password)


@pytest.mark.asyncio
async def test_no_token_obtained_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test Unable to obtain token for Envoy authentication"""
    version = "7.6.175_standard"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    with patch("pyenphase.EnvoyTokenAuth._obtain_token", return_value=None):
        envoy = Envoy("127.0.0.1", client=test_client_session)
        await envoy.setup()
        with pytest.raises(EnvoyAuthenticationError):
            await envoy.authenticate("username", "password")


@pytest.mark.asyncio
async def test_jwt_failure_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test Unable to verify token for Envoy authentication"""
    version = "7.6.175_standard"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    from .common import override_mock

    override_mock(
        mock_aioresponse,
        "get",
        "https://127.0.0.1" + URL_AUTH_CHECK_JWT,
        status=404,
        body="no jwt",
    )

    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    with pytest.raises(EnvoyAuthenticationError):
        await envoy.authenticate("username", "password")


@pytest.mark.asyncio
async def test_no_remote_login_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test Unable to login to Enlighten to obtain session ID from https://enlighten.enphaseenergy.com/login/login.json?"""
    version = "7.6.175_standard"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    from .common import override_mock

    override_mock(
        mock_aioresponse,
        "post",
        "https://enlighten.enphaseenergy.com/login/login.json?",
        status=500,
        payload={
            "session_id": "1234567890",
            "user_id": "1234567890",
            "user_name": "test",
            "first_name": "Test",
            "is_consumer": True,
            "manager_token": "1234567890",
        },
    )
    override_mock(
        mock_aioresponse,
        "post",
        "https://entrez.enphaseenergy.com/tokens",
        status=500,
        body="token",
    )
    override_mock(
        mock_aioresponse,
        "get",
        "https://127.0.0.1/auth/check_jwt",
        status=200,
        payload={},
    )

    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    with pytest.raises(EnvoyAuthenticationError):
        await envoy.authenticate("username", "password")


@pytest.mark.asyncio
async def test_no_remote_token_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test Unable to obtain token for Envoy authentication from https://entrez.enphaseenergy.com/tokens"""
    version = "7.6.175_standard"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    from .common import override_mock

    # The login endpoint is already mocked with 200 by start_7_firmware_mock
    # Only override the tokens endpoint to fail
    override_mock(
        mock_aioresponse,
        "post",
        "https://entrez.enphaseenergy.com/tokens",
        status=500,
        body="token",
    )

    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    with pytest.raises(EnvoyAuthenticationError):
        await envoy.authenticate("username", "password")

    assert isinstance(envoy.auth, EnvoyTokenAuth)
    with pytest.raises(EnvoyAuthenticationRequired):
        assert envoy.auth.token_type == "owner"


@pytest.mark.asyncio
async def test_enlighten_json_error_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test Unable to decode response from Enlighten"""
    version = "7.6.175_standard"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    from .common import override_mock

    # Override the login endpoint to return invalid JSON
    override_mock(
        mock_aioresponse,
        "post",
        "https://enlighten.enphaseenergy.com/login/login.json?",
        status=200,
        body="nojson",
    )

    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    with pytest.raises(EnvoyAuthenticationError):
        await envoy.authenticate("username", "password")


@pytest.mark.asyncio
async def test_enlighten_json_missing_is_consumer_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test response from Enlighten is missing is_consumer and manager keys"""
    version = "7.6.175_standard"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    from .common import override_mock

    # Override the login endpoint to return missing is_consumer and manager keys in JSON
    override_mock(
        mock_aioresponse,
        "post",
        "https://enlighten.enphaseenergy.com/login/login.json?",
        status=200,
        payload={
            "session_id": "1234567890",
            "user_id": "1234567890",
            "user_name": "test",
            "first_name": "Test",
        },
    )

    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    await envoy.authenticate("username", "password")
    assert isinstance(envoy.auth, EnvoyTokenAuth)
    assert not envoy.auth.is_consumer
    assert envoy.auth.manager_token == ""


@pytest.mark.asyncio
async def test_enlighten_json_missing_session_id_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test response from Enlighten is missing session_id"""
    version = "7.6.175_standard"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    from .common import override_mock

    # Override the login endpoint to return missing session_id keys in JSON
    override_mock(
        mock_aioresponse,
        "post",
        "https://enlighten.enphaseenergy.com/login/login.json?",
        status=200,
        payload={
            "nosession_id": "1234567890",
            "user_id": "1234567890",
            "user_name": "test",
            "first_name": "Test",
        },
    )
    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    with pytest.raises(EnvoyAuthenticationError):
        await envoy.authenticate("username", "password")


@pytest.mark.asyncio
async def test_token_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test auth using token"""
    version = "7.6.175_standard"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()

    token = jwt.encode(
        payload={"name": "envoy", "exp": 1707837780, "enphaseUser": "owner"},
        key="useaverylongsecretofatleast32bytestoavoidajwtsecuritywarning",
        algorithm="HS256",
    )

    await envoy.authenticate("username", "password", token)
    assert isinstance(envoy.auth, EnvoyTokenAuth)
    assert envoy.auth.expire_timestamp == 1707837780
    assert envoy.auth.token == token
    assert envoy.auth.token_type == "owner"

    # test cookies function now cookies are not on request
    assert envoy.auth.cookies == {}

    # execute refresh code cov
    await envoy.auth.refresh()

    # cov: test force no serial error
    # Your firmware requires token authentication,
    # but no envoy serial number was provided to obtain the token
    used_serial = envoy.auth.envoy_serial
    envoy.auth.envoy_serial = None
    with pytest.raises(EnvoyAuthenticationError):
        await envoy.auth.refresh()
    envoy.auth.envoy_serial = used_serial

    # cov: test force no cloud credentials error
    # Your firmware requires token authentication
    # but no cloud credentials were provided to obtain the token
    envoy.auth.cloud_password = None
    with pytest.raises(EnvoyAuthenticationError):
        await envoy.auth.refresh()


@pytest.mark.asyncio
async def test_remote_login_response_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test enlighten login response for is_consumer and manager_token"""
    version = "7.6.175_standard"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    # set log level to info 1 time for GET and 1 time for POST to improve COV
    with temporary_log_level("pyenphase", logging.INFO):
        envoy = Envoy("127.0.0.1", client=test_client_session)
        await envoy.setup()
        await envoy.authenticate("username", "password")

    assert isinstance(envoy.auth, EnvoyTokenAuth)
    assert envoy.auth.manager_token == "1234567890"
    assert envoy.auth.is_consumer

    # read unused auth from EnvoyTokenAuth to improve COV
    assert envoy.auth.auth is None


@pytest.mark.asyncio
async def test_device_data_with_8_2_4345_with_device_data(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test device data fails with no auth"""
    version = "8.2.4345_with_device_data"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    from .common import override_mock, updater_features

    # 401 on device data endpoint and inverters production endpoint
    # This simulates the case where the user does not have access to any inverter data
    for url in [
        "https://127.0.0.1" + URL_DEVICE_DATA,
        "http://127.0.0.1" + URL_DEVICE_DATA,
        "https://127.0.0.1" + URL_PRODUCTION_INVERTERS,
        "http://127.0.0.1" + URL_PRODUCTION_INVERTERS,
    ]:
        override_mock(
            mock_aioresponse,
            "get",
            url,
            status=401,
            body="no device data",
        )

    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    await envoy.authenticate("username", "password")

    await envoy.probe()
    assert SupportedFeatures.INVERTERS not in envoy.supported_features

    # 200 on device data endpoint only
    # This simulates the case where the user has access to device data but not to inverter production data
    for url in [
        "https://127.0.0.1" + URL_DEVICE_DATA,
        "http://127.0.0.1" + URL_DEVICE_DATA,
    ]:
        override_mock(
            mock_aioresponse,
            "get",
            url,
            status=200,
            payload=await load_json_fixture(version, "ivp_pdm_device_data"),
        )

    await envoy.probe()
    assert updater_features(envoy._updaters) == {
        "EnvoyDeviceDataInvertersUpdater": SupportedFeatures.INVERTERS
        | SupportedFeatures.DETAILED_INVERTERS,
        "EnvoyEnembleUpdater": SupportedFeatures.ENCHARGE | SupportedFeatures.ENPOWER,
        "EnvoyMetersUpdater": SupportedFeatures.CTMETERS,
        "EnvoyProductionJsonUpdater": SupportedFeatures.METERING
        | SupportedFeatures.TOTAL_CONSUMPTION
        | SupportedFeatures.NET_CONSUMPTION
        | SupportedFeatures.PRODUCTION,
        "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
    }


# S



"""Python wrapper for Enphase Envoy API."""

from .auth import EnvoyTokenAuth
from .envoy import AUTH_TOKEN_MIN_VERSION, Envoy, register_updater
from .exceptions import (
    EnvoyAuthenticationError,
    EnvoyAuthenticationRequired,
    EnvoyCommunicationError,
    EnvoyError,
    EnvoyFirmwareCheckError,
    EnvoyFirmwareFatalCheckError,
    EnvoyProbeFailed,
)
from .models.acb import EnvoyACBPower, EnvoyBatteryAggregate
from .models.c6combiner import EnvoyC6CC
from .models.collar import EnvoyCollar
from .models.dry_contacts import EnvoyDryContactSettings, EnvoyDryContactStatus
from .models.encharge import EnvoyEncharge, EnvoyEnchargeAggregate, EnvoyEnchargePower
from .models.enpower import EnvoyEnpower
from .models.envoy import EnvoyData
from .models.inverter import EnvoyInverter
from .models.system_consumption import EnvoySystemConsumption
from .models.system_production import EnvoySystemProduction
from .models.tariff import EnvoyTariff

__all__ = (
    AUTH_TOKEN_MIN_VERSION,
    "register_updater",
    "Envoy",
    "EnvoyData",
    "EnvoyTokenAuth",
    "EnvoyError",
    "EnvoyCommunicationError",
    "EnvoyFirmwareCheckError",
    "EnvoyFirmwareFatalCheckError",
    "EnvoyAuthenticationError",
    "EnvoyAuthenticationRequired",
    "EnvoyProbeFailed",
    "EnvoyInverter",
    "EnvoySystemConsumption",
    "EnvoySystemProduction",
    "EnvoyEncharge",
    "EnvoyEnchargeAggregate",
    "EnvoyEnchargePower",
    "EnvoyEnpower",
    "EnvoyACBPower",
    "EnvoyBatteryAggregate",
    "EnvoyDryContactSettings",
    "EnvoyDryContactStatus",
    "EnvoyCollar",
    "EnvoyC6CC",
    "EnvoyTariff",
)



"""Test specific envoy firmware issues post v7."""

import logging

import aiohttp
import pytest
from aioresponses import aioresponses

from pyenphase.envoy import UPDATERS, Envoy, SupportedFeatures, register_updater
from pyenphase.exceptions import EnvoyAuthenticationRequired
from pyenphase.updaters.api_v1_production_inverters import (
    EnvoyApiV1ProductionInvertersUpdater,
)
from pyenphase.updaters.device_data_inverters import EnvoyDeviceDataInvertersUpdater

from .common import (
    get_mock_envoy,
    override_mock,
    prep_envoy,
    start_7_firmware_mock,
    updater_features,
)

LOGGER = logging.getLogger(__name__)


@pytest.mark.parametrize(
    (
        "version",
        "part_number",
        "updaters",
        "watts_now",
        "watt_hours_today",
        "watt_hours_last_7_days",
        "watt_hours_lifetime",
    ),
    [
        (
            "8.2.4264_metered_noct",
            "800-00554-r03",
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyProductionJsonFallbackUpdater": SupportedFeatures.PRODUCTION,
                "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
            },
            5521,
            70,
            1521,
            32465106,
        ),
        (
            "7.6.114_without_cts",
            "800-00656-r06",
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyApiV1ProductionUpdater": SupportedFeatures.PRODUCTION,
            },
            586,
            10363,
            101742,
            1544282,
        ),
        (
            "7.3.466_metered_disabled_cts",
            "800-00654-r08",
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyProductionJsonFallbackUpdater": SupportedFeatures.PRODUCTION,
                "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
            },
            751,
            4425,
            111093,
            702919,
        ),
    ],
    ids=[
        "8.2.4264_metered_noct",
        "7.6.114_without_cts",
        "7.3.466_metered_disabled_cts",
    ],
)
@pytest.mark.asyncio
async def test_metered_noct(
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
    version: str,
    part_number: str,
    updaters: dict[str, SupportedFeatures],
    caplog: pytest.LogCaptureFixture,
    watts_now: int,
    watt_hours_today: int,
    watt_hours_last_7_days: int,
    watt_hours_lifetime: int,
) -> None:
    """Verify metered without CT production data with pre and post 8.2.4264 firmware."""
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)
    caplog.set_level(logging.DEBUG)

    envoy = await get_mock_envoy(test_client_session)
    data = envoy.data
    assert data is not None

    assert updater_features(envoy._updaters) == updaters
    assert envoy.part_number == part_number
    assert envoy.phase_count == 1

    assert not data.system_consumption
    assert envoy.ct_meter_count == 0
    assert envoy.phase_mode is None
    assert envoy.consumption_meter_type is None
    assert not data.system_consumption_phases
    assert not data.system_production_phases
    assert data.system_production is not None
    assert data.system_production.watts_now == watts_now
    assert data.system_production.watt_hours_today == watt_hours_today
    assert data.system_production.watt_hours_last_7_days == watt_hours_last_7_days
    assert data.system_production.watt_hours_lifetime == watt_hours_lifetime


@pytest.mark.asyncio
async def test_multiple_inverter_sources(
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
) -> None:
    """Test that multiple inverters from different sources are handled correctly."""
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", "8.2.4345_with_device_data")

    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    await envoy.authenticate("username", "password")

    # Preserve the original updaters
    original_updaters = UPDATERS.copy()

    # Remove existing inverter updaters
    UPDATERS[:] = [
        updater
        for updater in UPDATERS
        if updater
        not in (EnvoyApiV1ProductionInvertersUpdater, EnvoyDeviceDataInvertersUpdater)
    ]

    # Add the inverter production endpoint updater followed by the device data updater
    prod_remover = register_updater(EnvoyApiV1ProductionInvertersUpdater)
    device_data_remover = register_updater(EnvoyDeviceDataInvertersUpdater)

    # Verify that the production updater is used first
    await envoy.probe()
    assert updater_features(envoy._updaters) == {
        "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
        "EnvoyEnembleUpdater": SupportedFeatures.ENCHARGE | SupportedFeatures.ENPOWER,
        "EnvoyMetersUpdater": SupportedFeatures.CTMETERS,
        "EnvoyProductionJsonUpdater": SupportedFeatures.METERING
        | SupportedFeatures.TOTAL_CONSUMPTION
        | SupportedFeatures.NET_CONSUMPTION
        | SupportedFeatures.PRODUCTION,
        "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
    }

    # Remove both updaters and re-add them in reverse order
    prod_remover()
    device_data_remover()
    device_data_remover = register_updater(EnvoyDeviceDataInvertersUpdater)
    prod_remover = register_updater(EnvoyApiV1ProductionInvertersUpdater)

    # Verify that the device data updater is used first
    await envoy.probe()
    assert updater_features(envoy._updaters) == {
        "EnvoyDeviceDataInvertersUpdater": SupportedFeatures.INVERTERS
        | SupportedFeatures.DETAILED_INVERTERS,
        "EnvoyEnembleUpdater": SupportedFeatures.ENCHARGE | SupportedFeatures.ENPOWER,
        "EnvoyMetersUpdater": SupportedFeatures.CTMETERS,
        "EnvoyProductionJsonUpdater": SupportedFeatures.METERING
        | SupportedFeatures.TOTAL_CONSUMPTION
        | SupportedFeatures.NET_CONSUMPTION
        | SupportedFeatures.PRODUCTION,
        "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
    }

    # Restore the original updaters
    UPDATERS.clear()
    for updater in original_updaters:
        register_updater(updater)


@pytest.mark.parametrize(
    "version",
    [
        "8.2.4264_metered_noct",
        "7.6.114_without_cts",
        "7.3.466_metered_disabled_cts",
    ],
    ids=[
        "8.2.4264_metered_noct",
        "7.6.114_without_cts",
        "7.3.466_metered_disabled_cts",
    ],
)
@pytest.mark.asyncio
async def test_client_session_close(
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
    version: str,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test client session close code COV."""
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)
    caplog.set_level(logging.DEBUG)

    # pass aiohttp client session for envoy to use
    envoy = await get_mock_envoy(client_session=test_client_session)
    data = envoy.data
    assert data is not None
    assert envoy._client is not None
    assert not envoy._client.closed
    await envoy.close()
    # it's our client, pyenphase will not close it on close
    assert not envoy._client.closed

    # test with pyenphase internal created client
    envoy2 = await get_mock_envoy(client_session=None)
    data = envoy2.data
    assert data is not None
    assert envoy2._client is not None
    assert not envoy2._client.closed
    await envoy2.close()
    # it's pyenphase's client, will close it on close
    assert envoy2._client.closed

    envoy3 = await get_mock_envoy(client_session=None)
    data = envoy3.data
    assert data is not None
    assert not envoy3._client.closed

    # force close internal envoy client for cov test
    await envoy3._client.close()
    assert envoy3._client.closed
    await envoy3.close()
    # was closed already, should still be closed.
    assert envoy3._client.closed


@pytest.mark.parametrize(
    "version",
    [
        "7.3.130_no_consumption",
    ],
    ids=[
        "7.3.130_no_consumption",
    ],
)
@pytest.mark.asyncio
async def test_early_v7_with_all_401(
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
    version: str,
) -> None:
    """Test early v7 startup where several probe endpoints return auth-required."""
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    # endpoints return 401 on early v7 firmwares
    for endpoint in [
        "production.json?details=1",
        "production",
        "ivp/pdm/device_data",
        "api/v1/production/inverters",
        "ivp/ensemble/inventory",
        "admin/lib/tariff",
    ]:
        override_mock(
            mock_aioresponse,
            "get",
            f"https://127.0.0.1/{endpoint}",
            exception=EnvoyAuthenticationRequired("Test early v7 401"),
            repeat=True,
            payload=[],
        )

    envoy = await get_mock_envoy(client_session=test_client_session)
    assert updater_features(envoy._updaters) == {
        "EnvoyApiV1ProductionUpdater": SupportedFeatures.PRODUCTION,
        "EnvoyMetersUpdater": SupportedFeatures.DUALPHASE | SupportedFeatures.CTMETERS,
    }



import logging
from typing import Any

from ..const import URL_PRODUCTION_INVERTERS, SupportedFeatures
from ..exceptions import ENDPOINT_PROBE_EXCEPTIONS, EnvoyAuthenticationRequired
from ..models.envoy import EnvoyData
from ..models.inverter import EnvoyInverter
from .base import EnvoyUpdater

_LOGGER = logging.getLogger(__name__)


class EnvoyApiV1ProductionInvertersUpdater(EnvoyUpdater):
    """Class to handle updates for inverter production data."""

    async def probe(
        self, discovered_features: SupportedFeatures
    ) -> SupportedFeatures | None:
        """Probe the Envoy for this updater and return SupportedFeatures."""
        if SupportedFeatures.INVERTERS in discovered_features:
            # Already discovered from another updater
            return None

        try:
            await self._json_probe_request(URL_PRODUCTION_INVERTERS)
        except ENDPOINT_PROBE_EXCEPTIONS as e:
            _LOGGER.debug(
                "Production endpoint not found at %s: %s", URL_PRODUCTION_INVERTERS, e
            )
            return None
        except EnvoyAuthenticationRequired as e:
            _LOGGER.debug(
                "Disabling inverters production endpoint as user does"
                " not have access to %s: %s",
                URL_PRODUCTION_INVERTERS,
                e,
            )
            return None

        self._supported_features |= SupportedFeatures.INVERTERS
        return self._supported_features

    async def update(self, envoy_data: EnvoyData) -> None:
        """Update the Envoy for this updater."""
        inverters_data: list[dict[str, Any]] = await self._json_request(
            URL_PRODUCTION_INVERTERS
        )
        envoy_data.raw[URL_PRODUCTION_INVERTERS] = inverters_data
        envoy_data.inverters = {
            inverter["serialNumber"]: EnvoyInverter.from_v1_api(inverter)
            for inverter in inverters_data
        }



"""Test tenacity retry functioning."""

import asyncio
import logging
from typing import Any

import aiohttp
import pytest
from aioresponses import aioresponses
from tenacity import wait_none

from pyenphase import Envoy
from pyenphase.const import (
    DEFAULT_MAX_REQUEST_ATTEMPTS,
    MAX_PROBE_REQUEST_ATTEMPTS,
)
from pyenphase.exceptions import (
    EnvoyAuthenticationRequired,
    EnvoyCommunicationError,
    EnvoyFirmwareCheckError,
    EnvoyFirmwareFatalCheckError,
    EnvoyHTTPStatusError,
)

from .common import load_fixture, override_mock, prep_envoy, start_7_firmware_mock


# Helper function to create ClientConnectorError
def _make_client_connector_error(msg="Test error"):
    """Create a ClientConnectorError that can be converted to string."""

    # Create a simple mock object with the minimal attributes needed
    class MockConnKey:
        ssl = True
        host = "127.0.0.1"
        port = 443

    return aiohttp.ClientConnectorError(
        connection_key=MockConnKey(), os_error=OSError(msg)
    )


LOGGER = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_full_connected_from_start_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test envoy connected and replying from start"""
    version = "7.6.175_standard"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    envoy = Envoy("127.0.0.1", client=test_client_session)
    # remove the waits between retries for this test
    # retries are defined by MAX_PROBE_REQUEST_ATTEMPTS and MAX_PROBE_REQUEST_DELAY
    envoy._firmware._get_info.retry.wait = wait_none()

    await envoy.setup()
    await envoy.authenticate("username", "password")

    # Ensure that there was 1 attempt only.
    stats: dict[str, Any] = envoy._firmware._get_info.statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == 1

    assert envoy.firmware == "7.6.175"
    assert envoy.part_number == "800-00656-r06"

    data = await envoy.update()
    assert data


@pytest.mark.asyncio
async def test_full_disconnected_from_start_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test envoy disconnect at start, should return EnvoyFirmwareFatalCheckError."""
    start_7_firmware_mock(mock_aioresponse)
    envoy = Envoy("127.0.0.1", client=test_client_session)
    # remove the waits between retries for this test and set known retries
    # retries are defined by MAX_PROBE_REQUEST_ATTEMPTS and MAX_PROBE_REQUEST_DELAY
    envoy._firmware._get_info.retry.wait = wait_none()

    # Mock both HTTPS and HTTP since firmware code falls back to HTTP
    mock_aioresponse.get(
        "https://127.0.0.1/info",
        exception=_make_client_connector_error("Test timeoutexception"),
        repeat=True,
    )
    mock_aioresponse.get(
        "http://127.0.0.1/info",
        exception=_make_client_connector_error("Test timeoutexception"),
        repeat=True,
    )

    with pytest.raises(
        EnvoyFirmwareFatalCheckError, match="Unable to connect to Envoy"
    ):
        await envoy.setup()

    # Ensure there were MAX_PROBE_REQUEST_ATTEMPTS attempts.
    stats: dict[str, Any] = envoy._firmware._get_info.statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == MAX_PROBE_REQUEST_ATTEMPTS


@pytest.mark.asyncio
async def test_2_timeout_from_start_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test envoy timeout at start, timeout is not in retry loop."""
    start_7_firmware_mock(mock_aioresponse)
    envoy = Envoy("127.0.0.1", client=test_client_session)
    # remove the waits between retries for this test and set known retries
    # retries are defined by MAX_PROBE_REQUEST_ATTEMPTS and MAX_PROBE_REQUEST_DELAY
    envoy._firmware._get_info.retry.wait = wait_none()

    # Mock both HTTPS and HTTP since firmware code falls back to HTTP
    mock_aioresponse.get(
        "https://127.0.0.1/info",
        exception=asyncio.TimeoutError("Test timeoutexception"),
    )
    mock_aioresponse.get(
        "http://127.0.0.1/info", exception=asyncio.TimeoutError("Test timeoutexception")
    )

    with pytest.raises(
        EnvoyFirmwareFatalCheckError, match="Timeout connecting to Envoy"
    ):
        await envoy.setup()

    # Ensure there was only 1 attempt
    stats: dict[str, Any] = envoy._firmware._get_info.statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == 1


@pytest.mark.asyncio
async def test_httperror_from_start_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test envoy httperror at start, is not in retry loop."""
    start_7_firmware_mock(mock_aioresponse)
    # Don't call prep_envoy because we want to control the /info response

    envoy = Envoy("127.0.0.1", client=test_client_session)
    # remove the waits between retries for this test and set known retries
    # retries are defined by MAX_PROBE_REQUEST_ATTEMPTS and MAX_PROBE_REQUEST_DELAY
    envoy._firmware._get_info.retry.wait = wait_none()

    # The test expects no retries, which means we need to trigger the code path
    # that doesn't retry. Since _get_info retries all https exceptions on http, we need to
    # make the first https attempt succeeds but return bad data that causes setup() to fail
    mock_aioresponse.get(
        "https://127.0.0.1/info",
        status=500,  # Return HTTP error status
        body="Server Error",
    )

    with pytest.raises(EnvoyFirmwareCheckError, match="500"):
        await envoy.setup()

    # Ensure there was 1 attempt.
    stats: dict[str, Any] = envoy._firmware._get_info.statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == 1


@pytest.mark.asyncio
async def test_1_timeout_from_start_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test envoy timeout at start, timeout is not in retry loop but tries http after https."""
    version = "7.6.175_standard"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    envoy = Envoy("127.0.0.1", client=test_client_session)
    # remove the waits between retries for this test and set known retries
    # retries are defined by MAX_PROBE_REQUEST_ATTEMPTS and MAX_PROBE_REQUEST_DELAY
    envoy._firmware._get_info.retry.wait = wait_none()

    # Mock both HTTPS and HTTP since firmware code falls back to HTTP
    mock_aioresponse.get(
        "https://127.0.0.1/info",
        exception=asyncio.TimeoutError("Test timeoutexception"),
    )
    mock_aioresponse.get(
        "http://127.0.0.1/info", status=200, body=await load_fixture(version, "info")
    )

    await envoy.setup()
    await envoy.authenticate("username", "password")

    # Ensure there was 1 attempt.
    stats: dict[str, Any] = envoy._firmware._get_info.statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == 1

    assert envoy.firmware == "7.6.175"
    assert envoy.part_number == "800-00656-r06"

    data = await envoy.update()
    assert data


@pytest.mark.asyncio
async def test_5_not_connected_at_start_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test 5 connection failures at start and last one works"""
    version = "7.6.175_standard"
    start_7_firmware_mock(mock_aioresponse)
    # Don't call prep_envoy because we want to control the /info response

    envoy = Envoy("127.0.0.1", client=test_client_session)
    # remove the waits between retries for this test and set known retries
    # retries are defined by MAX_PROBE_REQUEST_ATTEMPTS and MAX_PROBE_REQUEST_DELAY
    envoy._firmware._get_info.retry.wait = wait_none()

    # Each retry attempt tries HTTPS first, then falls back to HTTP
    # We want 2 full failures (4 requests) then success on the 3rd attempt (request 5-6)
    # Attempt 1: HTTPS fails, HTTP fails
    mock_aioresponse.get(
        "https://127.0.0.1/info",
        exception=_make_client_connector_error("Test timeoutexception"),
    )
    mock_aioresponse.get(
        "http://127.0.0.1/info",
        exception=_make_client_connector_error("Test timeoutexception"),
    )
    # Attempt 2: HTTPS fails, HTTP fails
    mock_aioresponse.get(
        "https://127.0.0.1/info",
        exception=_make_client_connector_error("Test timeoutexception"),
    )
    mock_aioresponse.get(
        "http://127.0.0.1/info",
        exception=_make_client_connector_error("Test timeoutexception"),
    )
    # Attempt 3: HTTPS fails, HTTP succeeds
    mock_aioresponse.get(
        "https://127.0.0.1/info",
        exception=_make_client_connector_error("Test timeoutexception"),
    )
    mock_aioresponse.get(
        "http://127.0.0.1/info", status=200, body=await load_fixture(version, "info")
    )
    await envoy.setup()
    await envoy.authenticate("username", "password")

    # Ensure that there were retries.
    stats: dict[str, Any] = envoy._firmware._get_info.statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == 3

    assert envoy.firmware == "7.6.175"
    assert envoy.part_number == "800-00656-r06"

    # Now set up the other endpoints for the update call
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    data = await envoy.update()
    assert data


@pytest.mark.asyncio
async def test_2_network_errors_at_start_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test 2 network error failures at start and 3th works"""
    version = "7.6.175_standard"
    start_7_firmware_mock(mock_aioresponse)
    # Don't call prep_envoy because we want to control the /info response

    envoy = Envoy("127.0.0.1", client=test_client_session)
    # remove the waits between retries for this test and set known retries
    # retries are defined by MAX_PROBE_REQUEST_ATTEMPTS and MAX_PROBE_REQUEST_DELAY
    envoy._firmware._get_info.retry.wait = wait_none()

    # we need 2 side effects for each try as https and then http is attempted
    mock_aioresponse.get(
        "https://127.0.0.1/info", exception=aiohttp.ClientError("Test timeoutexception")
    )
    mock_aioresponse.get(
        "https://127.0.0.1/info",
        exception=_make_client_connector_error("Test timeoutexception"),
    )
    mock_aioresponse.get(
        "https://127.0.0.1/info", status=200, body=await load_fixture(version, "info")
    )

    await envoy.setup()
    await envoy.authenticate("username", "password")

    # Ensure that there were retries.
    stats: dict[str, Any] = envoy._firmware._get_info.statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == 3

    assert envoy.firmware == "7.6.175"
    assert envoy.part_number == "800-00656-r06"

    # Now set up the other endpoints for the update call
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    data = await envoy.update()
    assert data


@pytest.mark.asyncio
async def test_network_errors_at_start_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test network error failures at start"""
    start_7_firmware_mock(mock_aioresponse)
    # Don't call prep_envoy because we want to control the /info response

    envoy = Envoy("127.0.0.1", client=test_client_session)
    # remove the waits between retries for this test and set known retries
    # retries are defined by MAX_PROBE_REQUEST_ATTEMPTS and MAX_PROBE_REQUEST_DELAY
    envoy._firmware._get_info.retry.wait = wait_none()

    # We need 3 failures, each could try HTTPS then HTTP fallback
    mock_aioresponse.get(
        "https://127.0.0.1/info",
        exception=aiohttp.ClientError("Test timeoutexception"),
        repeat=True,
    )

    with pytest.raises(
        EnvoyFirmwareCheckError, match="Unable to query firmware version"
    ):
        await envoy.setup()

    # Ensure that there were retries.
    stats: dict[str, Any] = envoy._firmware._get_info.statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == MAX_PROBE_REQUEST_ATTEMPTS


@pytest.mark.asyncio
async def test_noconnection_at_probe_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test 3 network error failures at start"""
    version = "7.6.175_standard"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    envoy = Envoy("127.0.0.1", client=test_client_session)
    # remove the waits between retries for this test and set known retries
    # retries are defined by MAX_PROBE_REQUEST_ATTEMPTS and MAX_PROBE_REQUEST_DELAY
    envoy._firmware._get_info.retry.wait = wait_none()
    envoy.probe_request.retry.wait = wait_none()

    await envoy.setup()
    await envoy.authenticate("username", "password")

    # Ensure there was 1 attempt.
    stats: dict[str, Any] = envoy._firmware._get_info.statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == 1

    override_mock(
        mock_aioresponse,
        "get",
        "https://127.0.0.1/admin/lib/tariff",
        exception=aiohttp.ClientError("Test timeoutexception"),
        repeat=True,
    )

    await envoy.probe()

    stats = envoy.probe_request.statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == MAX_PROBE_REQUEST_ATTEMPTS


@pytest.mark.asyncio
async def test_noconnection_at_update_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test network error failures at update"""
    version = "7.6.175_standard"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    envoy = Envoy("127.0.0.1", client=test_client_session)
    # remove the waits between retries for this test and set known retries
    # retries are defined by MAX_PROBE_REQUEST_ATTEMPTS and MAX_PROBE_REQUEST_DELAY
    envoy._firmware._get_info.retry.wait = wait_none()
    envoy.set_retry_policy(wait_multiplier=0)

    await envoy.setup()
    await envoy.authenticate("username", "password")

    # Ensure there was 1 attempt.
    stats: dict[str, Any] = envoy._firmware._get_info.statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == 1

    await envoy.probe()

    stats = envoy.probe_request.statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == 1

    # Test timeout exceptions - need to override existing mock
    override_mock(
        mock_aioresponse,
        "get",
        "https://127.0.0.1/api/v1/production",
        exception=asyncio.TimeoutError("Test timeoutexception"),
        repeat=True,
    )

    # Clear endpoint cache to force retries
    envoy._endpoint_cache.clear()

    with pytest.raises(EnvoyCommunicationError, match="Timeout"):
        await envoy.update()

    # Don't check statistics here - they get reset between update() calls

    # Test connection errors
    envoy._endpoint_cache.clear()
    override_mock(
        mock_aioresponse,
        "get",
        "https://127.0.0.1/api/v1/production",
        exception=_make_client_connector_error("Test timeoutexception"),
        repeat=True,
    )

    with pytest.raises(EnvoyCommunicationError, match="aiohttp ClientError"):
        await envoy.update()

    # Check statistics immediately after the failed update
    stats = envoy.last_request_statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == DEFAULT_MAX_REQUEST_ATTEMPTS

    # Test general client errors (equivalent to RemoteProtocolError)
    envoy._endpoint_cache.clear()
    override_mock(
        mock_aioresponse,
        "get",
        "https://127.0.0.1/api/v1/production",
        exception=aiohttp.ClientError("Test timeoutexception"),
        repeat=True,
    )

    with pytest.raises(EnvoyCommunicationError, match="aiohttp ClientError"):
        await envoy.update()

    stats = envoy.last_request_statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == DEFAULT_MAX_REQUEST_ATTEMPTS

    # Test network errors (using ClientConnectorError as equivalent)
    envoy._endpoint_cache.clear()
    override_mock(
        mock_aioresponse,
        "get",
        "https://127.0.0.1/api/v1/production",
        exception=_make_client_connector_error("Test timeoutexception"),
        repeat=True,
    )

    with pytest.raises(EnvoyCommunicationError, match="aiohttp ClientError"):
        await envoy.update()

    stats = envoy.last_request_statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == DEFAULT_MAX_REQUEST_ATTEMPTS

    # other error EnvoyAuthenticationRequired should end cycle
    # First mock will be consumed, then the EnvoyAuthenticationRequired will stop retries
    envoy._endpoint_cache.clear()
    override_mock(
        mock_aioresponse,
        "get",
        "https://127.0.0.1/api/v1/production",
        exception=_make_client_connector_error("Test timeoutexception"),
    )
    # We can't directly mock EnvoyAuthenticationRequired from aioresponses,
    # so we'll use a 401 status to trigger it
    mock_aioresponse.get(
        "https://127.0.0.1/api/v1/production",
        status=401,
        payload={"message": "Test authentication required"},
    )
    mock_aioresponse.get(
        "https://127.0.0.1/api/v1/production",
        exception=_make_client_connector_error("Should not reach this"),
    )

    with pytest.raises(EnvoyAuthenticationRequired):
        await envoy.update()

    stats = envoy.last_request_statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == 2


@pytest.mark.asyncio
async def test_bad_request_status_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test request status not between 200-300."""
    version = "7.6.175_standard"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)
    envoy = Envoy("127.0.0.1", client=test_client_session)
    # remove the waits between retries for this test and set known retries
    # retries are defined by MAX_PROBE_REQUEST_ATTEMPTS and MAX_PROBE_REQUEST_DELAY
    envoy._firmware._get_info.retry.wait = wait_none()

    await envoy.setup()
    await envoy.authenticate("username", "password")

    data = await envoy.update()
    assert data

    # force status 503 on /api/vi/production
    # test status results in EnvoyHTTPStatusError
    override_mock(
        mock_aioresponse, "get", "https://127.0.0.1/api/v1/production", status=503
    )

    with pytest.raises(EnvoyHTTPStatusError, match="503"):
        await envoy.update()

    stats = envoy.last_request_statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == 1


@pytest.mark.asyncio
async def test_retry_policy(
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
) -> None:
    """Test envoy retry policy."""
    version = "7.6.175_with_cts"
    start_7_firmware_mock(mock_aioresponse)
    envoy = Envoy("127.0.0.1", client=test_client_session)

    # remove the waits between retries for this test and set known retries
    # retries are defined by MAX_PROBE_REQUEST_ATTEMPTS and MAX_PROBE_REQUEST_DELAY
    envoy._firmware._get_info.retry.wait = wait_none()

    # verify MAX_PROBE_REQUEST_ATTEMPTS are used in fw setup
    # Mock both HTTPS and HTTP since firmware code falls back to HTTP
    mock_aioresponse.get(
        "https://127.0.0.1/info",
        exception=_make_client_connector_error("Test timeoutexception"),
        repeat=True,
    )
    mock_aioresponse.get(
        "http://127.0.0.1/info",
        exception=_make_client_connector_error("Test timeoutexception"),
        repeat=True,
    )

    with pytest.raises(EnvoyFirmwareFatalCheckError):
        await envoy.setup()

    # verify probe attempts where used
    stats: dict[str, Any] = envoy._firmware._get_info.statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == MAX_PROBE_REQUEST_ATTEMPTS

    # restore info mock
    override_mock(
        mock_aioresponse,
        "get",
        "https://127.0.0.1/info",
        status=200,
        body=await load_fixture(version, "info"),
        repeat=True,
    )

    override_mock(
        mock_aioresponse,
        "get",
        "http://127.0.0.1/info",
        status=200,
        body=await load_fixture(version, "info"),
        repeat=True,
    )
    # setup all other mocks
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    await envoy.setup()
    await envoy.authenticate("username", "password")

    # do initial probe to get updaters identified
    await envoy.probe()

    # verify DEFAULT_MAX_REQUEST_ATTEMPTS are used for update
    override_mock(
        mock_aioresponse,
        "get",
        "https://127.0.0.1/ivp/meters",
        exception=asyncio.TimeoutError("Test timeoutexception"),
        repeat=True,
    )
    # remove waits between retries
    envoy.set_retry_policy(wait_multiplier=0)

    with pytest.raises(EnvoyCommunicationError):
        await envoy.update()

    # verify custom attempts where used
    stats = envoy.last_request_statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == DEFAULT_MAX_REQUEST_ATTEMPTS

    # Add second envoy instance
    mock_aioresponse.get("https://127.0.0.2/auth/check_jwt", status=200, repeat=True)
    envoy2 = Envoy("127.0.0.2", client=test_client_session)
    await prep_envoy(mock_aioresponse, "127.0.0.2", version)
    await envoy2.setup()
    await envoy2.authenticate("username", "password")

    # disable wait between retries
    envoy2.set_retry_policy(wait_multiplier=0)
    data2 = await envoy2.update()
    assert data2

    # set retry policy to use custom retries on first envoy
    envoy.set_retry_policy(max_delay=600, max_attempts=8)
    with pytest.raises(EnvoyCommunicationError):
        await envoy.update()

    # verify custom attempts where used
    stats = envoy.last_request_statistics
    assert "attempt_number" in stats
    assert stats["attempt_number"] == 8

    # verify default retries are used on second envoy
    override_mock(
        mock_aioresponse,
        "get",
        "https://127.0.0.2/ivp/meters",
        exception=asyncio.TimeoutError("Test timeoutexception"),
        repeat=True,
    )
    with pytest.raises(EnvoyCommunicationError):
        await envoy2.update()

    # verify default attempts were used on second envoy as well
    stats2: dict[str, Any] = envoy2.last_request_statistics
    assert "attempt_number" in stats2
    assert stats2["attempt_number"] == DEFAULT_MAX_REQUEST_ATTEMPTS



"""Model for an Enphase microinverter."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass(slots=True)
class EnvoyInverter:
    """Model for an Enphase microinverter."""

    serial_number: str
    last_report_date: int
    last_report_watts: int
    max_report_watts: int
    dc_voltage: float | None = None
    dc_current: float | None = None
    ac_voltage: float | None = None
    ac_current: float | None = None
    ac_frequency: float | None = None
    temperature: float | None = None
    lifetime_energy: int | None = None
    energy_produced: float | None = None
    energy_today: int | None = None
    last_report_duration: int | None = None

    @classmethod
    def from_v1_api(cls, data: dict[str, Any]) -> EnvoyInverter:
        """Initialize from the V1 API."""
        return cls(
            serial_number=data["serialNumber"],
            last_report_date=data["lastReportDate"],
            last_report_watts=data["lastReportWatts"],
            max_report_watts=data["maxReportWatts"],
        )

    @classmethod
    def from_device_data(cls, data: dict[str, Any]) -> EnvoyInverter:
        """Initialize from device data."""

        def safe_convert_milli(value: float | None) -> float | None:
            return value / 1000.0 if value is not None else None

        # if these don't exist they raise KeyError to not use this model but use from_v1_api
        channel = data["channels"][0]
        last_reading = channel["lastReading"]

        # these four are minimal data set, if one fails keyerror will raise
        serial_number = data["sn"]
        last_report_date = last_reading["endDate"]
        last_report_watts = channel["watts"]["now"]
        max_report_watts = channel["watts"]["max"]

        # get data to avoid divide errors if None
        duration = last_reading.get("duration")
        period_joules_produced = last_reading.get("joulesProduced")
        lifetime = channel.get("lifetime")
        lifetime_joulesProduced = lifetime.get("joulesProduced") if lifetime else None
        watthours = channel.get("wattHours")

        return cls(
            serial_number=serial_number,
            last_report_date=last_report_date,
            last_report_watts=last_report_watts,
            max_report_watts=max_report_watts,
            # next ones may return none as they didn't exist before in the model
            dc_voltage=safe_convert_milli(last_reading.get("dcVoltageINmV")),
            dc_current=safe_convert_milli(last_reading.get("dcCurrentINmA")),
            ac_voltage=safe_convert_milli(last_reading.get("acVoltageINmV")),
            ac_current=safe_convert_milli(last_reading.get("acCurrentInmA")),
            ac_frequency=safe_convert_milli(last_reading.get("acFrequencyINmHz")),
            temperature=last_reading.get("channelTemp"),
            lifetime_energy=round(lifetime_joulesProduced / 3600.0)
            if lifetime_joulesProduced is not None
            else None,
            energy_produced=round(period_joules_produced / duration / 3.6, 3)
            if period_joules_produced is not None and duration is not None
            else None,
            energy_today=watthours.get("today") if watthours else None,
            last_report_duration=duration,
        )



"""Model for the IQ Meter Collar."""

# Data Source: URL_ENSEMBLE_INVENTORY

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

# Required keys for IQ Meter Collar inventories
COLLAR_REQUIRED_KEYS: frozenset[str] = frozenset(
    {
        "admin_state",
        "admin_state_str",
        "communicating",
        "img_load_date",
        "img_pnum_running",
        "installed",
        "last_rpt_date",
        "part_num",
        "serial_num",
        "temperature",
        "mid_state",
        "grid_state",
        "control_error",
        "collar_state",
    }
)


@dataclass(slots=True)
class EnvoyCollar:
    """Model for the Enphase IQ Meter Collar."""

    admin_state: int
    admin_state_str: str
    firmware_loaded_date: int
    firmware_version: str
    installed_date: int
    last_report_date: int
    communicating: bool
    mid_state: str
    grid_state: str
    part_number: str
    serial_number: str
    temperature: int
    temperature_unit: str
    control_error: int
    collar_state: str

    @classmethod
    def from_api(cls, inventory: dict[str, Any]) -> EnvoyCollar | None:
        """Initialize from the API. Returns None if required keys are missing."""
        missing_keys = COLLAR_REQUIRED_KEYS - set(inventory)
        if missing_keys:
            return None
        return cls(
            admin_state=inventory["admin_state"],
            admin_state_str=inventory["admin_state_str"],
            communicating=inventory["communicating"],
            firmware_loaded_date=inventory["img_load_date"],
            firmware_version=inventory["img_pnum_running"],
            installed_date=inventory["installed"],
            last_report_date=inventory["last_rpt_date"],
            part_number=inventory["part_num"],
            serial_number=inventory["serial_num"],
            temperature=inventory["temperature"],
            temperature_unit="C",
            mid_state=inventory["mid_state"],
            grid_state=inventory["grid_state"],
            control_error=inventory["control_error"],
            collar_state=inventory["collar_state"],
        )



"""Test firmware functions."""

import logging

import aiohttp
import pytest
from aioresponses import aioresponses

from pyenphase import Envoy
from pyenphase.exceptions import EnvoyFirmwareCheckError

LOGGER = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_firmware_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test firmware is processed ok."""
    info = (
        "<?xml version='1.0' encoding='UTF-8'?>"
        "<envoy_info>"
        "  <device>"
        "    <sn>123456789012</sn>"
        "    <pn>800-12345-r99</pn>"
        "    <software>D7.8.901</software>"
        "  </device>"
        "</envoy_info>"
    )
    mock_aioresponse.get("https://127.0.0.1/info", status=200, body=info)
    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    assert envoy.firmware == "7.8.901"
    assert envoy.serial_number == "123456789012"
    assert envoy.part_number == "800-12345-r99"


@pytest.mark.asyncio
async def test_firmware_no_sn_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test missing serial number in info"""
    info = (
        "<?xml version='1.0' encoding='UTF-8'?>"
        "<envoy_info>"
        "  <device>"
        "    <pn>800-12345-r99</pn>"
        "    <software>D7.8.901</software>"
        "  </device>"
        "</envoy_info>"
    )
    mock_aioresponse.get("https://127.0.0.1/info", status=200, body=info)
    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    assert envoy.firmware == "7.8.901"
    assert envoy.serial_number is None
    assert envoy.part_number == "800-12345-r99"


@pytest.mark.asyncio
async def test_firmware_no_pn_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test missing pb in info"""
    info = (
        "<?xml version='1.0' encoding='UTF-8'?>"
        "<envoy_info>"
        "  <device>"
        "    <sn>123456789012</sn>"
        "    <software>D7.8.901</software>"
        "  </device>"
        "</envoy_info>"
    )
    mock_aioresponse.get("https://127.0.0.1/info", status=200, body=info)
    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    assert envoy.firmware == "7.8.901"
    assert envoy.serial_number == "123456789012"
    assert envoy.part_number is None


@pytest.mark.asyncio
async def test_firmware_no_fw_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test missing fw in info"""
    info = (
        "<?xml version='1.0' encoding='UTF-8'?>"
        "<envoy_info>"
        "  <device>"
        "    <sn>123456789012</sn>"
        "    <pn>800-12345-r99</pn>"
        "  </device>"
        "</envoy_info>"
    )
    mock_aioresponse.get("https://127.0.0.1/info", status=200, body=info)
    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    assert not envoy.firmware
    assert envoy.serial_number == "123456789012"
    assert envoy.part_number == "800-12345-r99"


@pytest.mark.asyncio
async def test_firmware_no_device_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test missing device xml segment in info"""
    info = "<?xml version='1.0' encoding='UTF-8'?><envoy_info></envoy_info>"
    mock_aioresponse.get("https://127.0.0.1/info", status=200, body=info)
    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    assert not envoy.firmware
    assert envoy.serial_number is None
    assert envoy.part_number is None


@pytest.mark.asyncio
async def test_firmware_no_200__with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test other status as 200 returned"""
    info = "<?xml version='1.0' encoding='UTF-8'?><envoy_info></envoy_info>"
    mock_aioresponse.get("https://127.0.0.1/info", status=500, body=info)
    envoy = Envoy("127.0.0.1", client=test_client_session)
    with pytest.raises(EnvoyFirmwareCheckError):
        await envoy.setup()


@pytest.mark.asyncio
async def test_firmware_metered_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test firmware is processed ok."""
    info = (
        "<?xml version='1.0' encoding='UTF-8'?>"
        "<envoy_info>"
        "  <device>"
        "    <sn>123456789012</sn>"
        "    <pn>800-12345-r99</pn>"
        "    <software>D7.8.901</software>"
        "    <imeter>true</imeter>"
        "  </device>"
        "</envoy_info>"
    )
    mock_aioresponse.get("https://127.0.0.1/info", status=200, body=info)
    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    assert envoy.is_metered


@pytest.mark.asyncio
async def test_firmware_not_metered_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test firmware is processed ok."""
    info = (
        "<?xml version='1.0' encoding='UTF-8'?>"
        "<envoy_info>"
        "  <device>"
        "    <sn>123456789012</sn>"
        "    <pn>800-12345-r99</pn>"
        "    <software>D7.8.901</software>"
        "    <imeter>false</imeter>"
        "  </device>"
        "</envoy_info>"
    )
    mock_aioresponse.get("https://127.0.0.1/info", status=200, body=info)
    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    assert not envoy.is_metered


@pytest.mark.asyncio
async def test_firmware_missing_metered_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test firmware is processed ok."""
    info = (
        "<?xml version='1.0' encoding='UTF-8'?>"
        "<envoy_info>"
        "  <device>"
        "    <sn>123456789012</sn>"
        "    <pn>800-12345-r99</pn>"
        "    <software>D7.8.901</software>"
        "  </device>"
        "</envoy_info>"
    )
    mock_aioresponse.get("https://127.0.0.1/info", status=200, body=info)
    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    assert not envoy.is_metered



"""Envoy Firmware detection"""

import asyncio
import logging
import time

import aiohttp
from awesomeversion import AwesomeVersion
from lxml import etree  # nosec
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    stop_after_delay,
    wait_random_exponential,
)

from .const import LOCAL_TIMEOUT, MAX_PROBE_REQUEST_ATTEMPTS, MAX_PROBE_REQUEST_DELAY
from .exceptions import EnvoyFirmwareCheckError, EnvoyFirmwareFatalCheckError

_LOGGER = logging.getLogger(__name__)


class EnvoyFirmware:
    """Class for querying and determining the Envoy firmware version."""

    __slots__ = (
        "_client",
        "_firmware_version",
        "_host",
        "_metered",
        "_part_number",
        "_serial_number",
        "_url",
    )

    def __init__(
        self,
        _client: aiohttp.ClientSession,
        host: str,
    ) -> None:
        """
        Class for querying and determining the Envoy firmware version.

        :param client: aiohttp ClientSession not verifying SSL
            certificates, see :class:`pyenphase.ssl`.
        :param host: Envoy DNS name or IP address
        """
        self._client = _client
        self._host = host
        self._firmware_version: str | None = None
        self._serial_number: str | None = None
        self._part_number: str | None = None
        self._url: str = ""
        self._metered: bool = False

    @retry(
        retry=retry_if_exception_type(aiohttp.ClientError),
        wait=wait_random_exponential(multiplier=2, max=5),
        stop=stop_after_delay(MAX_PROBE_REQUEST_DELAY)
        | stop_after_attempt(MAX_PROBE_REQUEST_ATTEMPTS),
        reraise=True,
    )
    async def _get_info(self) -> tuple[int, bytes]:
        """
        Perform GET request to /info endpoint on envoy.

        Try GET request to https://<host>/info to read info endpoint.
        If connection error or timeout, retry on http://<host>/info.

        Will retry up to :any:`MAX_PROBE_REQUEST_ATTEMPTS` times
        or :any:`MAX_PROBE_REQUEST_DELAY` elapsed at next try, which
        ever comes first on network or remote protocol errors.
        HTTP status is not verified.

        :return: tuple of (status_code, content)
        """
        self._url = f"https://{self._host}/info"
        _LOGGER.debug("Requesting %s with timeout %s", self._url, LOCAL_TIMEOUT)
        try:
            resp = await self._client.get(self._url, timeout=LOCAL_TIMEOUT)
            return resp.status, await resp.read()
        except (aiohttp.ClientConnectorError, asyncio.TimeoutError):
            # Firmware < 7.0.0 does not support HTTPS so we need to try HTTP
            # as a fallback, worse sometimes http will redirect to https://localhost
            # which is not helpful
            self._url = f"http://{self._host}/info"
            _LOGGER.debug("Retrying to %s with timeout %s", self._url, LOCAL_TIMEOUT)
            resp = await self._client.get(self._url, timeout=LOCAL_TIMEOUT)
            return resp.status, await resp.read()

    async def setup(self) -> None:
        """
        Obtain the firmware version, serial-number and part-number from Envoy.

        Read /info on Envoy, accessible without authentication.
        Store firmware version, serial-number and part-number properties
        from xml response.

        Reads first on HTTPS, if that fails on HTTP for firmware < 7.
        Will retry up to :any:`MAX_PROBE_REQUEST_ATTEMPTS` times
        or :any:`MAX_PROBE_REQUEST_DELAY` elapsed at next try, which
        ever comes first on network or remote protocol errors.

        .. code-block:: python

            connector = aiohttp.TCPConnector(ssl=create_no_verify_ssl_context())
            client = aiohttp.ClientSession(connector=connector)
            firmware = EnvoyFirmware(client,host)
            await firmware.setup()
            print(firmware.version)

        :raises EnvoyFirmwareFatalCheckError: if connection or timeout
            failure occurs
        :raises EnvoyFirmwareCheckError: on http errors or any HTTP
            status other then 200
        """
        # <envoy>/info will return XML with the firmware version
        debugon = _LOGGER.isEnabledFor(logging.DEBUG)
        if debugon:
            request_start = time.monotonic()
        try:
            status_code, content = await self._get_info()
        except asyncio.TimeoutError as ex:
            raise EnvoyFirmwareFatalCheckError(
                500, "Timeout connecting to Envoy"
            ) from ex
        except aiohttp.ClientConnectorError as ex:
            raise EnvoyFirmwareFatalCheckError(
                500, "Unable to connect to Envoy"
            ) from ex
        except aiohttp.ClientError as ex:
            raise EnvoyFirmwareCheckError(
                500, "Unable to query firmware version"
            ) from ex

        if status_code == 200:
            if debugon:
                request_end = time.monotonic()
                _LOGGER.debug(
                    "Request reply in %s sec from %s status %s: %s",
                    round(request_end - request_start, 1),
                    self._url,
                    status_code,
                    content,
                )
            xml = etree.fromstring(content)  # nosec
            if (device_tag := xml.find("device")) is not None:
                if (software_tag := device_tag.find("software")) is not None:
                    self._firmware_version = AwesomeVersion(
                        software_tag.text[1:]
                    )  # need to strip off the leading 'R' or 'D'
                if (sn_tag := device_tag.find("sn")) is not None:
                    self._serial_number = sn_tag.text
                if (pn_tag := device_tag.find("pn")) is not None:
                    self._part_number = pn_tag.text
                if (imeter_tag := device_tag.find("imeter")) is not None:
                    self._metered = imeter_tag.text == "true"
                return

        else:
            # If we get a different status code, raise an exception
            raise EnvoyFirmwareCheckError(status_code, content.decode())

    @property
    def version(self) -> AwesomeVersion:
        """
        Return firmware version as read from Envoy.

        :return: Envoy firmware version or None if
            :class:`pyenphase.firmware.EnvoyFirmware.setup` was not used
        """
        return self._firmware_version

    @property
    def serial(self) -> str | None:
        """
        Return serial number as read from Envoy.

        :return: Envoy serial number or None if
            :class:`pyenphase.firmware.EnvoyFirmware.setup` was not used
        """
        return self._serial_number

    @property
    def part_number(self) -> str | None:
        """
        Return part number as read from Envoy.

        :return: Envoy part number or None if
            :class:`pyenphase.firmware.EnvoyFirmware.setup` was not used
        """
        return self._part_number

    @property
    def is_metered(self) -> bool:
        """
        Return imetered setting as read from Envoy.

        :return: Envoy info imetered setting. Only True if read and set in info

        """
        return self._metered



"""pyenphase constant definitions"""

import enum

import aiohttp
from awesomeversion import AwesomeVersion

# Versions
LEGACY_ENVOY_VERSION = AwesomeVersion("3.9.0")
ENSEMBLE_MIN_VERSION = AwesomeVersion("5.0.0")
AUTH_TOKEN_MIN_VERSION = AwesomeVersion("7.0.0")
METERED_NOCT_FALLBACK_TO_INVERTERS = AwesomeVersion("8.2.4264")

# Include in docs from here

# System Production
URL_PRODUCTION_INVERTERS = "/api/v1/production/inverters"
URL_DEVICE_DATA = "/ivp/pdm/device_data"
URL_PRODUCTION_V1 = "/api/v1/production"
URL_PRODUCTION_JSON = "/production.json?details=1"
URL_PRODUCTION = "/production"

# Authentication
URL_AUTH_CHECK_JWT = "/auth/check_jwt"

# Battery and Enpower Status
URL_DRY_CONTACT_STATUS = "/ivp/ensemble/dry_contacts"
URL_DRY_CONTACT_SETTINGS = "/ivp/ss/dry_contact_settings"
URL_ENSEMBLE_INVENTORY = "/ivp/ensemble/inventory"
URL_ENSEMBLE_STATUS = "/ivp/ensemble/status"
URL_ENSEMBLE_SECCTRL = "/ivp/ensemble/secctrl"
URL_ENCHARGE_BATTERY = "/ivp/ensemble/power"
URL_GRID_RELAY = "/ivp/ensemble/relay"
URL_POWER_EXPORT = "/uvp/ss/pel_settings"
URL_TARIFF = "/admin/lib/tariff"

# Generator Configuration
URL_GEN_CONFIG = "/ivp/ss/gen_config"
URL_GEN_MODE = "/ivp/ss/gen_mode"
URL_GEN_SCHEDULE = "/ivp/ss/gen_schedule"

# Meters data
ENDPOINT_URL_METERS = "/ivp/meters"
ENDPOINT_URL_METERS_READINGS = "/ivp/meters/readings"

# Interface configuration
ENDPOINT_URL_HOME = "/home"

# Include in docs to here

LOCAL_TIMEOUT = aiohttp.ClientTimeout(
    # The envoy can be slow to respond but fast to connect to we
    # need to set a long timeout for the read and a short timeout
    # for the connect
    total=45.0,
    connect=10.0,
    sock_read=45.0,
)  #: default timeout used for requests

# Requests should no longer retry after max delay (sec) or attempts since first try
# Default is LOCAL_TIMEOUT of 45 sec with some wait time between retries
# which results in retries after 45-90-135-180. max delay of 120 == 3 tries, 150 == 4
DEFAULT_MAX_REQUEST_DELAY = 150  #: default maximum request retry time in seconds
DEFAULT_MAX_REQUEST_ATTEMPTS = 6  #: default maximum request retry attempts
MAX_REQUEST_DELAY = DEFAULT_MAX_REQUEST_DELAY  #: deprecated alias
MAX_REQUEST_ATTEMPTS = DEFAULT_MAX_REQUEST_ATTEMPTS  #: deprecated alias
MAX_PROBE_REQUEST_DELAY = 50  #: maximum elapsed probe retry time in seconds
MAX_PROBE_REQUEST_ATTEMPTS = 4  #: maximum request probe retry attempts


class SupportedFeatures(enum.IntFlag):
    """
    Features available from Envoy

    Each supported feature maps to a specific data set or information
    that can be provided by an Envoy. Depending on actual make, firmware
    and installed components an Envoy may provide 1 or more features.
    All Envoy should at least report solar production, marked as PRODUCTION.

    Class :any:`EnvoyUpdater` updaters will set these features flags
    during the :any:`Envoy.probe` phase. During data collection by
    :any:`Envoy.update` each updater with set features will be used to
    collect the specific data.

    .. code-block:: python

        from pyenphase.const import SupportedFeatures

        # set METERING flag
        features |= SupportedFeatures.METERING

        # test features
        if features.PRODUCTION in supported_features:
            pass

        if features & SupportedFeatures.DUALPHASE:
            pass
    """

    INVERTERS = 1  #: Envoy reports solar panel inverters
    METERING = 2  #: Envoy reports active production meter
    TOTAL_CONSUMPTION = 4  #: Envoy reports total consumption
    NET_CONSUMPTION = 8  #: Envoy reports net consumption
    ENCHARGE = 16  #: Envoy reports encharge data
    ENPOWER = 32  #: Envoy reports Enpower data
    PRODUCTION = 64  #: Envoy reports solar production data
    TARIFF = 128  #: Envoy reports tariff information
    DUALPHASE = 256  #: Envoy metered is configured in split phase mode
    THREEPHASE = 512  #: Envoy metered is configured in three phase mode
    CTMETERS = 1024  #: Envoy has enabled CT meter(s)
    GENERATOR = 2048  #: Envoy reports generator data
    ACB = 4096  #: Envoy reports ACB Battery data
    DETAILED_INVERTERS = 8192  #: Detailed inverter data is reported
    COLLAR = 0x4000  #: Envoy reports a Collar
    C6CC = 0x8000  #: Envoy reports a C6 Combiner controller


class PhaseNames(enum.StrEnum):
    """Electricity grid phase names."""

    PHASE_1 = "L1"  #: first phase (1, A, ..)
    PHASE_2 = "L2"  #: second phase (2, B, ..)
    PHASE_3 = "L3"  #: third phase (3, C, ..)


#: list to access :any:`PhaseNames` by numerical index.
#:
#: .. code-block:: python
#:
#:     phase_count = 2
#:     for phase in range(phase_count):
#:         print(production[PHASENAMES[phase]])
#:
PHASENAMES: list[str] = list(PhaseNames)



from abc import abstractmethod
from collections.abc import Awaitable, Callable
from typing import Any

import aiohttp
from awesomeversion import AwesomeVersion

from ..const import SupportedFeatures
from ..exceptions import EnvoyHTTPStatusError
from ..json import json_loads
from ..models.common import CommonProperties
from ..models.envoy import EnvoyData


class EnvoyUpdater:
    """Base class for Envoy updaters."""

    def __init__(
        self,
        envoy_version: AwesomeVersion,
        probe_request: Callable[[str], Awaitable[aiohttp.ClientResponse]],
        request: Callable[[str], Awaitable[aiohttp.ClientResponse]],
        common_properties: CommonProperties,
    ) -> None:
        """
        Base class for Envoy updaters.

        Updaters should implement a subclass of EnvoyUpdater.

        .. code-block:: python

            class EnvoyXyzUpdater(EnvoyUpdater):

        :param envoy_version: firmware version Envoy is running
        :param probe_request: callable specified by
            :any:`Envoy` to send probe request to the Envoy
            during :any:`Envoy.probe`
        :param request: callable specified by :any:`Envoy` to
            send request to the Envoy during :any:`Envoy.update`
        :param common_properties: properties to share between
            probe and update or between updaters
        """
        self._envoy_version = envoy_version
        self._probe_request = probe_request
        self._request = request
        self._supported_features = SupportedFeatures(0)
        self._common_properties = common_properties

    async def _json_request(self, end_point: str) -> Any:
        """
        Make a request to the Envoy and return the JSON response.

        Updaters should use this to collect data during :any:`update` method.

        .. code-block:: python

            xyz_json: dict[str, Any] = await self._json_request(
                "/xyz/endpoint"
            )

        :param end_point: Envoy endpoint to request. See :any:`Envoy.request`
        :raises EnvoyHTTPStatusError: If http status not in 2xx range
        :return: JSON content from response
        :seealso: :any:`Envoy.request`
        """
        response = await self._request(end_point)
        if not (200 <= response.status < 300):
            raise EnvoyHTTPStatusError(response.status, str(response.url))
        return json_loads(end_point, await response.read())

    async def _json_probe_request(self, end_point: str) -> Any:
        """
        Make a probe request to the Envoy and return the JSON response.

        Updaters should use this to collect data during :any:`probe` method.

        .. code-block:: python

            xyz_json: dict[str, Any] = await self._json_probe_request(
                "/xyz/endpoint"
            )

        :param end_point: Envoy endpoint to request. See :any:`Envoy.probe`
        :raises EnvoyHTTPStatusError: If http status not in 2xx range
        :return: JSON content from response
        :seealso: :any:`Envoy.probe_request`
        """
        response = await self._probe_request(end_point)
        if not (200 <= response.status < 300):
            raise EnvoyHTTPStatusError(response.status, str(response.url))
        return json_loads(end_point, await response.read())

    @abstractmethod
    async def probe(
        self, discovered_features: SupportedFeatures
    ) -> SupportedFeatures | None:
        """
        Probe the Envoy for this updater and return SupportedFeatures.

        Updater must implement a probe method to report which features it supports.
        Probe method of each registered updater is called by :any:`Envoy.probe`.
        Intent of probe is to determine if this updater should be used
        to obtain the data in its scope from the Envoy. If the Envoy
        model does not provide the data the method should return None.

        If the Envoy model does provide the data in scope, the method should
        return a supported features mask for which it can provide data.
        Any supported feature that is already included in the passed
        discovered_features parameter should NOT be included in the result
        as these are already provided by another updater and this updater
        should back off from providing that feature data.

        .. code-block:: python

            async def probe(
                    self, discovered_features: SupportedFeatures
                ) -> SupportedFeatures | None:
                \"\"\"Probe the Envoy for this endpoint and return SupportedFeatures.\"\"\"
                if SupportedFeatures.MY_FEATURE in discovered_features:
                    # Already discovered from another updater
                    return None
                try:
                    json_data: dict[str, Any] = await self._json_probe_request(MY_ENDPOINT)
                except ENDPOINT_PROBE_EXCEPTIONS as e:
                    _LOGGER.debug("Endpoint not found at %s: %s", MY_ENDPOINT, e)
                    return None
                # process data
                return SupportedFeatures.MY_FEATURE

        :param discovered_features: Mask of already discovered SupportedFeatures
        :return: Mask of SupportedFeatures to be added to already discovered
            features (and not yet in discovered_features).
        """

    @abstractmethod
    async def update(self, envoy_data: EnvoyData) -> None:
        """
        Get data from the Envoy and store in EnvoyData.

        Updater must implement an update method to add/update data in
        :any:`EnvoyData`. Update method of each registered updater is
        called by :any:`Envoy.update`.

        The update method is expected to obtain the required data from
        the envoy, map it to the internal data model and store the data.
        It should also store retrieved raw data in :any:`EnvoyData.raw`

        .. code-block:: python

            async def update(self, envoy_data: EnvoyData) -> None:
                \"\"\"Get data from the Envoy and store in EnvoyData.\"\"\"
                json_data = await self._json_request(self.end_point)
                envoy_data.raw[self.end_point] = json_data
                # store applicable data in envoy_data

        :param envoy_data: Envoy data model to store collected data in
        """



"""Model for the Enpower dry contact relays."""

# Data Source: URL_DRY_CONTACT_SETTINGS (primary) & URL_DRY_CONTACT_STATUS

from __future__ import annotations

from dataclasses import dataclass
from enum import StrEnum
from typing import Any


class DryContactStatus(StrEnum):
    OPEN = "open"
    CLOSED = "closed"


class DryContactAction(StrEnum):
    APPLY = "apply"
    SHED = "shed"
    SCHEDULE = "schedule"
    NONE = "none"


class DryContactType(StrEnum):
    NONE = "NONE"
    PV = "PV"
    LOAD = "LOAD"
    THIRD_PARTY_PV = "3RD-PV"


class DryContactMode(StrEnum):
    MANUAL = "manual"
    STATE_OF_CHARGE = "soc"


@dataclass(slots=True)
class EnvoyDryContactStatus:
    """Model for the Enpower dry contact relay status."""

    id: str
    status: str

    @classmethod
    def from_api(cls, relay: dict[str, Any]) -> EnvoyDryContactStatus:
        """Initialize from the API."""
        return cls(
            id=relay["id"],
            status=DryContactStatus(relay["status"]),
        )


@dataclass(slots=True)
class EnvoyDryContactSettings:
    """Model for the Enpower dry contact relay settings."""

    id: str
    black_start: float | None
    essential_end_time: float | None
    essential_start_time: float | None
    generator_action: DryContactAction
    grid_action: DryContactAction
    load_name: str
    manual_override: bool | None
    micro_grid_action: DryContactAction
    mode: DryContactMode
    override: bool
    priority: float | None
    pv_serial_nb: list[Any]
    soc_high: float
    soc_low: float
    type: DryContactType

    @classmethod
    def from_api(cls, relay: dict[str, Any]) -> EnvoyDryContactSettings:
        """Initialize from the API."""
        return cls(
            id=relay["id"],
            black_start=relay.get("black_s_start"),
            essential_end_time=relay.get("essential_end_time"),
            essential_start_time=relay.get("essential_start_time"),
            generator_action=DryContactAction(relay["gen_action"]),
            grid_action=DryContactAction(relay["grid_action"]),
            load_name=relay["load_name"],
            manual_override=(
                relay["manual_override"] == "true"
                if relay.get("manual_override")
                else None
            ),
            micro_grid_action=DryContactAction(relay["micro_grid_action"]),
            mode=DryContactMode(relay["mode"]),
            override=relay["override"] == "true",
            priority=relay.get("priority"),
            pv_serial_nb=relay["pv_serial_nb"],
            soc_high=relay["soc_high"],
            soc_low=relay["soc_low"],
            type=DryContactType(relay["type"]),
        )

    def to_api(self) -> dict[str, Any]:
        """Convert to API format."""
        retval = {
            "id": self.id,
            "gen_action": self.generator_action,
            "grid_action": self.grid_action,
            "load_name": self.load_name,
            # boolean values must be passed to the API as a lowercase string
            "manual_override": str(self.manual_override).lower(),
            "micro_grid_action": self.micro_grid_action,
            "mode": self.mode,
            "override": str(self.override).lower(),
            "pv_serial_nb": self.pv_serial_nb,
            "soc_high": self.soc_high,
            "soc_low": self.soc_low,
            "type": self.type,
        }

        if self.black_start is not None:
            retval["black_s_start"] = self.black_start
        if self.essential_start_time is not None:
            retval["essential_start_time"] = self.essential_start_time
        if self.essential_end_time is not None:
            retval["essential_end_time"] = self.essential_end_time
        if self.priority is not None:
            retval["priority"] = self.priority
        if self.manual_override is not None:
            retval["manual_override"] = self.manual_override

        return retval



"""
Create test fixture file set for pyenphase by scanning known endpoints on Envoy.

execute python fixture_collector.py --help for directions

Copy this file to the Home Assistant config folder. Open a terminal on your HA system
Navigate to the config folder and execute python fixture_collector.py

Alternatively copy and execute this file to a system with pyenphase installed and
network connectivity to your Home Assistant config folder and the Envoy.
"""

import argparse
import asyncio
import contextlib
import getpass
import json
import logging
import os
import zipfile
from datetime import datetime

from aiohttp import ClientResponse

from pyenphase.envoy import Envoy
from pyenphase.exceptions import (
    EnvoyAuthenticationRequired,
    EnvoyFirmwareFatalCheckError,
)

# logging.basicConfig(level=logging.WARNING)

_LOGGER = logging.getLogger(__name__)

DEFAULT_ENDPOINTS = [
    "/info",
    "/api/v1/production",
    "/api/v1/production/inverters",
    "/production.json",
    "/production.json?details=1",
    "/production",
    "/ivp/ensemble/power",
    "/ivp/ensemble/inventory",
    "/ivp/ensemble/dry_contacts",
    "/ivp/ensemble/status",
    "/ivp/ensemble/secctrl",
    "/ivp/ss/dry_contact_settings",
    "/admin/lib/tariff",
    "/ivp/ss/gen_config",
    "/ivp/ss/gen_schedule",
    "/ivp/sc/pvlimit",
    "/ivp/ss/pel_settings",
    "/ivp/ensemble/generator",
    "/ivp/meters",
    "/ivp/meters/readings",
    "/ivp/pdm/device_data",
    "/home",
    "/inventory.json?deleted=1",
    "/inv",
    "/inventory",
    "/ivp/pdm/energy",
    "/ivp/meters/pvReading",
    "/ivp/meters/gridReading",
    "/ivp/meters/reports",
]


async def main(
    envoy_address: str | None,
    username: str | None,
    password: str | None,
    token: str | None,
    verbose: bool = False,
    label: str = "",
    clean: bool = False,
    endpoint_to_get: list[str] | None = None,
) -> None:
    host_arg = envoy_address
    env_host = os.environ.get("ENVOY_HOST")
    host = host_arg or env_host or "envoy.local"
    envoy = Envoy(host)

    try:
        await envoy.setup()
    except EnvoyFirmwareFatalCheckError as err:
        print(f"Could not connect to Envoy: {err.status_code} {err.status}")
        await envoy.close()
        return

    try:
        await envoy.authenticate(username=username, password=password, token=token)
    except EnvoyAuthenticationRequired:
        print("Could not authenticate with Envoy")
        await envoy.close()
        return

    target_dir = f"enphase-{envoy.firmware}{label}"
    with contextlib.suppress(FileExistsError):
        os.mkdir(target_dir)
        if verbose:
            print(f"Created folder: {target_dir}")

    end_points = endpoint_to_get or DEFAULT_ENDPOINTS

    assert envoy.auth  # nosec

    for end_point in end_points:
        # url = envoy.auth.get_endpoint_url(end_point)
        if verbose:
            print(f"Reading: {end_point}")
        try:
            start_time: datetime = datetime.now()
            response: ClientResponse = await envoy.request(end_point)
        except Exception as ex:
            _LOGGER.debug("Error getting %s", end_point, exc_info=ex)
            continue
        try:
            response_text = await response.text()
            end_time: datetime = datetime.now()
            duration_seconds = round((end_time - start_time).total_seconds(), 3)
            if verbose:
                print(f"{end_point} reply text read in: {duration_seconds}")
        except Exception as ex:
            _LOGGER.debug("Error getting %s", end_point, exc_info=ex)
            continue
        file_name = (
            end_point[1:]
            .replace("/", "_")
            .replace("?", "_")
            .replace("=", "_")
            .replace("&", "_")
            .replace(" ", "_")
        )
        file_path = os.path.join(target_dir, file_name)
        with open(file_path, "w", encoding="utf-8") as fixture_file:
            fixture_file.write(response_text)
            if verbose:
                print(f"Creating: {fixture_file.name}")

        with open(
            os.path.join(target_dir, f"{file_name}_log.json"), "w", encoding="utf-8"
        ) as metadata_file:
            # Remove potentially sensitive headers from being persisted to disk
            sensitive = {
                "authorization",
                "cookie",
                "set-cookie",
                "x-auth-token",
                "x-csrf-token",
                "x-api-key",
            }
            safe_headers = {
                k: v for k, v in response.headers.items() if k.lower() not in sensitive
            }
            metadata_file.write(
                json.dumps(
                    {
                        "headers": dict(safe_headers),
                        "code": response.status,
                        "duration_seconds": duration_seconds,
                    }
                )
            )

    if not clean or verbose:
        print(f"Fixtures written to {target_dir}")

    zip_file_name = f"{target_dir}.zip"
    with zipfile.ZipFile(zip_file_name, "w") as zip_file:
        for file_name in os.listdir(target_dir):
            zip_file.write(os.path.join(target_dir, file_name), file_name)
            if clean:
                os.remove(os.path.join(target_dir, file_name))

    print(f"Zip file written to {zip_file_name}")

    if clean:
        try:
            os.rmdir(target_dir)
            if verbose:
                print(f"Removed {target_dir}")
        except OSError as err:
            print(f"Could not clean folder: {err.strerror}")
        except FileNotFoundError:
            pass

    await envoy.close()


def _read_ha_config(file_path: str) -> dict[str, list[str | None]]:
    result: dict[str, list[str | None]] = {}
    try:
        with open(file_path) as fp:
            content = json.load(fp)
    except (FileNotFoundError, ValueError):
        return result

    if content:
        for entry in content["data"]["entries"]:
            if entry["domain"] != "enphase_envoy" or entry["source"] == "ignore":
                continue
            data = entry["data"]
            unique_id = entry["unique_id"]
            result[unique_id] = [
                data["host"],
                data["username"],
                data["password"],
                data["token"],
            ]

    return result


if __name__ == "__main__":
    description = (
        "Scan Enphase Envoy for endpoint list usable for pyenphase test fixtures. \
        Creates output folder enphase-<firmware>[label] with results of scan.\
        Zips content of created folder into enphase-<firmware>[label].zip.\
        \
        Optionally collect specified endpoints only.\
        "
    )
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "-d", "--debug", help="Enable debug logging", action="store_true"
    )
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument(
        "-c",
        "--clean",
        help="Remove created folder, but keep zip file",
        action="store_true",
    )
    parser.add_argument(
        "-l", "--label", help="Label to append to output folder and zip file names"
    )
    parser.add_argument(
        "-r",
        "--readhaconfig",
        const=".",
        nargs="?",
        dest="ha_config_folder",
        help="Read envoyname, username, password and token from HA config folder.\
            Use -r path_to_ha_config_folder. Default is current folder.\
                Overrides any specified username, password and token.\
                Reads <path_to_ha_config_folder>/.storage/core.config_entries.\
            ",
    )
    parser.add_argument(
        "envoyname",
        nargs="?",
        default="envoy.local",
        help="Envoy Name or IP address. IP is preferred, default is envoy.local",
    )
    parser.add_argument(
        "-u", "--username", help="Username (for Envoy or for Enphase token website)"
    )
    parser.add_argument(
        "-p", "--password", help="Password (blank or for Enphase token website)"
    )
    parser.add_argument(
        "-t",
        "--token",
        help="Enphase owner token or @path_to_file to read token from file",
    )
    parser.add_argument(
        "-e",
        "--endpoint",
        help="Comma-separated list of endpoints to read (e.g. /info,/home). Endpoints start with /.",
    )

    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    host = args.envoyname
    username: str | None = args.username
    password: str | None = args.password
    read_ha_config: str = args.ha_config_folder
    verbose: bool = args.verbose
    endpoints: list[str] | None = (
        [ep.strip() for ep in args.endpoint.split(",") if ep.strip()]
        if args.endpoint
        else []
    )

    config_entries: dict[str, list[str | None]] = {}
    target_ha_file: str = ""

    if args.ha_config_folder:
        target_ha_file = os.path.join(read_ha_config, ".storage", "core.config_entries")
        config_entries = _read_ha_config(target_ha_file)
    else:
        username = args.username
        password = args.password
        token = args.token
        if not token:
            token = os.environ.get("ENVOY_TOKEN", getpass.getpass("Enter the token: "))
        if token and token[0] == "@":
            try:
                with open(token[1:]) as f:
                    token = f.read()
            except FileNotFoundError:
                token = None
        if not username and not token:
            username = os.environ.get("ENVOY_USERNAME", input("Enter the Username: "))
        if not password and not token:
            password = os.environ.get(
                "ENVOY_PASSWORD", getpass.getpass("Enter the Password: ")
            )
        config_entries.update({"": [host, username, password, token]})

    for sn, configs in config_entries.items():
        host, username, password, token = configs
        if verbose:
            print(f"Using {host} sn: {sn}, {target_ha_file}")

        asyncio.run(
            main(
                envoy_address=host,
                username=username,
                password=password,
                token=token,
                verbose=verbose,
                label=args.label or "",
                clean=args.clean,
                endpoint_to_get=endpoints,
            )
        )



"""Envoy authentication methods."""

from abc import abstractmethod, abstractproperty
from typing import Any, cast

import aiohttp
import jwt
import orjson
from tenacity import retry, retry_if_exception_type, wait_random_exponential

from .const import LOCAL_TIMEOUT, URL_AUTH_CHECK_JWT
from .exceptions import EnvoyAuthenticationError, EnvoyAuthenticationRequired
from .ssl import SSL_CONTEXT


class EnvoyAuth:
    """Base class for local Envoy authentication."""

    def __init__(self, host: str) -> None:
        """
        Base class for local Envoy authentication.

        :param host: local Envoy DNS name or IP Address

        """

    @abstractmethod
    async def setup(self, client: aiohttp.ClientSession) -> None:
        """
        Setup token based authentication with the local Envoy.

        Required for Envoy firmware >= 7.0

        :param client: an aiohttp ClientSession to communicate with the local Envoy,

        """

    @abstractproperty
    def cookies(self) -> dict[str, str]:
        """Return the Envoy cookie."""

    @abstractproperty
    def auth(self) -> aiohttp.DigestAuthMiddleware | None:
        """
        Setup Digest authentication for local Envoy.

        Required for Envoy firmware < 7.0

        """

    @abstractproperty
    def headers(self) -> dict[str, str]:
        """Return the auth headers for Envoy communication."""

    @abstractmethod
    def get_endpoint_url(self, endpoint: str) -> str:
        """
        Return the URL for the endpoint.

        :param endpoint: Envoy Endpoint to access, start with leading /

        :return: formatted full URL string
        """


class EnvoyTokenAuth(EnvoyAuth):
    """Class to authenticate with Envoy using Tokens."""

    # autodoc docstring is supplied from __init__

    JSON_LOGIN_URL = "https://enlighten.enphaseenergy.com/login/login.json?"
    TOKEN_URL = "https://entrez.enphaseenergy.com/tokens"  # nosec

    def __init__(
        self,
        host: str,
        cloud_username: str | None = None,
        cloud_password: str | None = None,
        envoy_serial: str | None = None,
        token: str | None = None,
    ) -> None:
        """
        Class to authenticate with Envoy using Tokens.

        Use with Envoy firmware 7.x and newer

        :param host: local Envoy DNS name or IP Address
        :param cloud_username: Enligthen Cloud username, required to obtain new
            token when token is not specified or expired, defaults to None
        :param cloud_password: Enligthen Cloud password, required to obtain new
            token when token is not specified or expired, defaults to None
        :param envoy_serial: Envoy serial number, required to obtain new
            token when token is not specified or expired, defaults to None
        :param token: Token to use with authentication, if not specified,
            one will be obtained from Enlighten cloud if username, password
            and serial are specified, defaults to None

        """
        self.host = host
        self.cloud_username = cloud_username
        self.cloud_password = cloud_password
        self.envoy_serial = envoy_serial
        self._token = token
        self._is_consumer: bool = False
        self._manager_token: str | None = None
        self._cookies: dict[str, str] = {}

    async def setup(self, client: aiohttp.ClientSession) -> None:
        """
        Setup token based authentication with the local Envoy

        If no token is specified, a token is obtained from Enlighten Cloud using
        specified username, password and serialnumber. With the specified or obtained
        token, validates the token with the local Envoy. New or updated token
        can be accessed using the token property. Token is not stored persistent,
        caller should store and specify token over restarts.

        :param client: an aiohttp ClientSession to communicate with the local Envoy,
        :raises EnvoyAuthenticationError: Authentication failed with the local Envoy
            or no token could be obtained from Enlighten cloud due to error,
            missing parameters or Enlighten account issue.

        """
        if not self._token:
            self._token = await self._obtain_token()

        # Verify we have adequate credentials
        if not self._token:
            raise EnvoyAuthenticationError(
                "Unable to obtain token for Envoy authentication."
            )

        await self._check_jwt(client)

    @retry(
        retry=retry_if_exception_type(aiohttp.ClientError),
        wait=wait_random_exponential(multiplier=2, max=3),
    )
    async def _check_jwt(self, client: aiohttp.ClientSession) -> None:
        """Check the JWT token for Envoy authentication."""
        async with client.get(
            f"https://{self.host}{URL_AUTH_CHECK_JWT}",
            headers={"Authorization": f"Bearer {self.token}"},
            timeout=LOCAL_TIMEOUT,
        ) as resp:
            if resp.status == 200:
                self._cookies = {k: v.value for k, v in resp.cookies.items()}
                return

        raise EnvoyAuthenticationError(
            "Unable to verify token for Envoy authentication."
        )

    async def _obtain_token(self) -> str:
        """Obtain the token for Envoy authentication."""
        # Raise if we don't have cloud credentials
        if not self.cloud_username or not self.cloud_password:
            raise EnvoyAuthenticationError(
                "Your firmware requires token authentication, "
                " but no cloud credentials were provided to obtain the token."
            )
        # Raise if we are missing the envoy serial number
        if not self.envoy_serial:
            raise EnvoyAuthenticationError(
                "Your firmware requires token authentication, "
                "but no envoy serial number was provided to obtain the token."
            )
        # We require a new client that checks SSL certs
        timeout = aiohttp.ClientTimeout(total=10)
        async with aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(ssl=SSL_CONTEXT), timeout=timeout
        ) as cloud_client:
            # Login to Enlighten to obtain a session ID
            response = await self._post_json_with_cloud_client(
                cloud_client,
                self.JSON_LOGIN_URL,
                data={
                    "user[email]": self.cloud_username,
                    "user[password]": self.cloud_password,
                },
            )
            if response.status != 200:
                text = await response.text()
                raise EnvoyAuthenticationError(
                    "Unable to login to Enlighten to obtain session ID from "
                    f"{self.JSON_LOGIN_URL}: "
                    f"{response.status}: {text}"
                )
            try:
                response_json = orjson.loads(await response.text())
            except orjson.JSONDecodeError as err:
                text = await response.text()
                raise EnvoyAuthenticationError(
                    "Unable to decode response from Enlighten: "
                    f"{response.status}: {text}"
                ) from err

            self._is_consumer = response_json.get("is_consumer", False)
            self._manager_token = response_json.get("manager_token")

            # Obtain the token
            try:
                response = await self._post_json_with_cloud_client(
                    cloud_client,
                    self.TOKEN_URL,
                    json={
                        "session_id": response_json["session_id"],
                        "serial_num": self.envoy_serial,
                        "username": self.cloud_username,
                    },
                )
            except KeyError as err:
                # if no session_id is returned from login we can't continue
                # this happens if MFA is enabled on Enlighten account
                raise EnvoyAuthenticationError(
                    "No session id in Enlighten login reply, disable Multi Factor Authentication if enabled, "
                    f"status: {response.status}, received {response_json.keys()}"
                ) from err

            if response.status != 200:
                text = await response.text()
                raise EnvoyAuthenticationError(
                    "Unable to obtain token for Envoy authentication from "
                    f"{self.TOKEN_URL}: "
                    f"{response.status}: {text}"
                )
            return await response.text()

    async def refresh(self) -> None:
        """
        Refresh the token for Envoy authentication.

        Retrieves a new token from the Enlighten cloud using
        specified username, password and Envoy serial number of
        the class object. Updated token can be accessed
        using the token property. Token is not stored persistent,
        caller should store it after refresh and specify token
        over restarts.

        """
        self._token = await self._obtain_token()

    @property
    def expire_timestamp(self) -> int:
        """
        Return the expiration time for the token.

        Owner useraccount type tokens are valid for a year
        while installer tokens are only valid for 12 hours.

        :return: epoch expiration time
        """
        jwt_payload = jwt.decode(self.token, options={"verify_signature": False})
        return cast(int, jwt_payload["exp"])

    @property
    def token_type(self) -> str:
        """
        Return the enphase user type for the token.

        Enlighten user accounts can be type 'owner' or 'installer'. Both
        have access to the envoy base data. Installer has access to more
        data and configuration setup.

        :raises: EnvoyAuthenticationRequired if no prior authentication was done
        :return: 'owner' or 'installer'
        """
        if not self._token:
            raise EnvoyAuthenticationRequired(
                "You must authenticate to the Envoy before inspecting token."
            )
        jwt_payload = jwt.decode(self.token, options={"verify_signature": False})
        return jwt_payload["enphaseUser"]

    @retry(
        retry=retry_if_exception_type(aiohttp.ClientError),
        wait=wait_random_exponential(multiplier=2, max=3),
    )
    async def _post_json_with_cloud_client(
        self,
        cloud_client: aiohttp.ClientSession,
        url: str,
        data: dict[str, Any] | None = None,
        json: dict[str, Any] | None = None,
    ) -> aiohttp.ClientResponse:
        """Post to the Envoy API with the cloud client."""
        return await cloud_client.post(url, json=json, data=data)

    @property
    def token(self) -> str:
        """
        Return token used with the Envoy.

        Returns the current token, either the original specified token,
        or the token obtained from the Enlighten cloud if not specified
        or the refreshed token at expiration.

        Will assert if no token was ever specified or obtained.

        :return: jwt token string
        """
        assert self._token is not None  # nosec
        return self._token

    @property
    def manager_token(self) -> str:
        """
        Return manager token returned in enligthen login json.

        This property is only available if a token has been requested
        from the Enlighten cloud. This is only the case if no token
        was specified, or a token refresh was requested. If a valid
        token with a future expiration time was specified this method
        will return an empty string.

        :return: token string
        """
        return self._manager_token if self._manager_token is not None else ""

    @property
    def cookies(self) -> dict[str, str]:
        """
        Return cookies returned during setup of the envoy.

        Cookies received from the local Envoy during setup and local
        jwt check are stored in the class, this method returns these.

        :return: cookies dict
        """
        return self._cookies

    @property
    def is_consumer(self) -> bool:
        """
        Return is_consumer state returned in enligthen login json

        This property is only available if a token has been requested
        from the Enlighten cloud. This is only the case if no token
        was specified, or a token refresh was requested. If a valid
        token with a future expiration time was specified no login was
        attempted and this method will return the default false.
        If an installer account was used it will return false as well.

        :return: true if enlighten login was performed and
            used credentials are for consumer account, otherwise false
        """
        return self._is_consumer

    @property
    def auth(self) -> None:
        """
        Digest authentication for local Envoy.

        Not used with token authentication. Placeholder
        for EnvoyAuth abstractproperty

        :return: None
        """
        return None

    @property
    def headers(self) -> dict[str, str]:
        """
        Return the authentication headers for Envoy communication.

        Token authorization with Envoy requires an Authorization header
        in Bearer format with token.

        :return: token authorization header
        """
        return {"Authorization": f"Bearer {self.token}"}

    def get_endpoint_url(self, endpoint: str) -> str:
        """
        Return the URL for the endpoint.

        :param endpoint: Envoy Endpoint to access, start with leading /

        :return: formatted https URL string
        """
        return f"https://{self.host}{endpoint}"


class EnvoyLegacyAuth(EnvoyAuth):
    """Class to authenticate with legacy Envoy using digest."""

    def __init__(self, host: str, username: str, password: str) -> None:
        """
        Class to authenticate with legacy Envoy using digest.

        Use with Envoy firmware before 7.0

        :param host: local Envoy DNS name or IP Address
        :param local_username: Username to access Envoy
        :param local_password: Password to access Envoy
        """
        self.host = host
        self.local_username = username
        self.local_password = password
        self._auth_middleware: aiohttp.DigestAuthMiddleware | None = None

    @property
    def auth(self) -> aiohttp.DigestAuthMiddleware | None:
        """
        Digest authentication for local Envoy.

        Creates DigestAuthMiddleware based on username and password.

        :return: DigestAuthMiddleware for local Envoy or None
            if username and/or password are not specified
        """
        if not self.local_username or not self.local_password:
            return None
        if self._auth_middleware is None:
            self._auth_middleware = aiohttp.DigestAuthMiddleware(
                self.local_username, self.local_password
            )
        return self._auth_middleware

    async def setup(self, client: aiohttp.ClientSession) -> None:
        """
        Setup authentication with the local Envoy

        DigestAuth does not use additional setup,
        placeholder for EnvoyAuth abstractpropery.

        :param client: ClientSession to communicate with local Envoy
        """
        # No setup required for legacy authentication

    @property
    def headers(self) -> dict[str, str]:
        """
        Return the headers needed for Envoy authentication.

        DigestAuth does not use authorization header. Placeholder
        for EnvoyAuth abstractproperty.

        :return: empty dict
        """
        return {}

    def get_endpoint_url(self, endpoint: str) -> str:
        """
        Return the URL for the endpoint.

        :param endpoint: Envoy Endpoint to access, start with leading /

        :return: formatted http URL string
        """
        return f"http://{self.host}{endpoint}"

    @property
    def cookies(self) -> dict[str, str]:
        """
        Return cookies returned during setup of the envoy.

        DigestAuth does not use cookies. Placeholder
        for EnvoyAuth abstractproperty.

        :return: empty dict
        """
        return {}



import logging
from typing import Any

from ..const import URL_DEVICE_DATA, SupportedFeatures
from ..exceptions import ENDPOINT_PROBE_EXCEPTIONS, EnvoyAuthenticationRequired
from ..models.envoy import EnvoyData
from ..models.inverter import EnvoyInverter
from .base import EnvoyUpdater

_LOGGER = logging.getLogger(__name__)


class EnvoyDeviceDataInvertersUpdater(EnvoyUpdater):
    """Class to handle updates for inverter device data."""

    def _filter_inverters(self, inverters_data: dict[str, Any]) -> dict[str, Any]:
        """Filter and return only PCU inverter devices."""
        return {
            inverter["sn"]: inverter
            for id, inverter in inverters_data.items()
            if id not in ("deviceCount", "deviceDataLimit")
            and inverter["devName"] == "pcu"
        }

    async def probe(
        self, discovered_features: SupportedFeatures
    ) -> SupportedFeatures | None:
        """Probe the Envoy for this updater and return SupportedFeatures."""
        if SupportedFeatures.INVERTERS in discovered_features:
            # Already discovered from another updater
            return None

        try:
            inverters_data = await self._json_probe_request(URL_DEVICE_DATA)
        except ENDPOINT_PROBE_EXCEPTIONS as e:
            _LOGGER.debug(
                "Device data endpoint not found at %s: %s", URL_DEVICE_DATA, e
            )
            return None
        except EnvoyAuthenticationRequired as e:
            _LOGGER.debug(
                "Disabling inverters device data endpoint as user does"
                " not have access to %s: %s",
                URL_DEVICE_DATA,
                e,
            )
            return None
        # make sure deviceCount did not reach deviceDataLimit,
        # if more inverters are actually installed they will not be included
        # if so fall back to inverter production page
        try:
            if inverters_data["deviceCount"] >= inverters_data["deviceDataLimit"]:
                _LOGGER.debug(
                    "Disabling inverters device data endpoint "
                    " as deviceCount reached  deviceDataLimit %s: %s - %s",
                    URL_DEVICE_DATA,
                    inverters_data["deviceCount"],
                    inverters_data["deviceDataLimit"],
                )
                return None
        except KeyError as e:
            # if doesn't have these keys, fall back to inverter production
            _LOGGER.debug(
                "Disabling inverters device data endpoint "
                " as not all data fields are present %s: %s",
                URL_DEVICE_DATA,
                e,
            )
            return None

        # verify minimal data set to replace inverter production data is present
        try:
            filtered_inverters = self._filter_inverters(inverters_data)
            _ = {
                sn: EnvoyInverter.from_device_data(inverter)
                for sn, inverter in filtered_inverters.items()
            }

        except KeyError as e:
            # if any inverter returned None there's something messed by json format, fall back to production
            _LOGGER.debug(
                "Disabling inverters device data endpoint "
                " as not all data fields are present %s: %s",
                URL_DEVICE_DATA,
                e,
            )
            return None

        self._supported_features |= (
            SupportedFeatures.INVERTERS | SupportedFeatures.DETAILED_INVERTERS
        )
        return self._supported_features

    async def update(self, envoy_data: EnvoyData) -> None:
        """Update the Envoy for this updater."""
        inverters_data: dict[str, Any] = await self._json_request(URL_DEVICE_DATA)
        envoy_data.raw[URL_DEVICE_DATA] = inverters_data
        filtered_inverters = self._filter_inverters(inverters_data)
        envoy_data.inverters = {
            sn: EnvoyInverter.from_device_data(inverter)
            for sn, inverter in filtered_inverters.items()
        }



"""Enphase Envoy exceptions."""

import json

import aiohttp


class EnvoyError(Exception):
    """Base class for Envoy exceptions."""


class EnvoyFirmwareCheckError(EnvoyError):
    """
    Exception raised when unable to query the Envoy firmware version.

    - http error when sending request to Envoy
    - Any http status code other then 200 received

    :param status_code: http status code
    :param status: Error status description
    """

    def __init__(self, status_code: int, status: str) -> None:
        self.status_code = status_code
        self.status = status


class EnvoyFirmwareFatalCheckError(EnvoyError):
    """
    Exception raised when we should not retry getting the Envoy firmware version.

    - aiohttp timeout or connection error when sending request to Envoy

    :param status_code: http status code
    :param status: Error status description
    """

    def __init__(self, status_code: int, status: str) -> None:
        self.status_code = status_code
        self.status = status


class EnvoyAuthenticationError(EnvoyError):
    """
    Exception raised when Envoy Authentication fails.

    - When a jwt token authentication failure occurs with the local Envoy.
    - When using token authentication and no cloud credentials or envoy serial are specified
    - When a failure occurs during obtaining a token from the Enlighten cloud

    :param status: Error status description
    """

    def __init__(self, status: str) -> None:
        self.status = status


class EnvoyAuthenticationRequired(EnvoyError):
    """
    Exception raised when authentication hasn't been setup.

    - When communication with Envoy was attempted without setting up authentication.
    - When neither token nor username and/or password are specified during authentication.

    :param status: Error status description
    """

    def __init__(self, status: str) -> None:
        self.status = status


class EnvoyHTTPStatusError(EnvoyError):
    """
    Exception raised when unable to query the Envoy.

    - HTTP Status of request not in 200 range.

    :param status_code: http status code
    :param status: Error status description
    :param url: failing url
    """

    def __init__(self, status_code: int, url: str) -> None:
        self.status_code = status_code
        self.url = url
        super().__init__(f"HTTP status error {url} {status_code}")


class EnvoyProbeFailed(EnvoyError):
    """Exception raised when the Envoy probe fails."""


class EnvoyCommunicationError(EnvoyError):
    """
    Exception raised when the Envoy communication fails.

    - aiohttp.ClientError error occurs.
    - asyncio.TimeoutError error occurs
    """


class EnvoyFeatureNotAvailable(EnvoyError):
    """
    Exception raised when the Envoy feature is not available.

    - When using go on/off grid and ENPOWER feature is not available in Envoy

    """


class EnvoyPoorDataQuality(EnvoyError):
    """
    Exception raised when data identifies known issues.

    - FW 3.x production values all zero at startup

    :param status: Error status description
    """

    def __init__(self, status: str) -> None:
        self.status = status


ENDPOINT_PROBE_EXCEPTIONS = (
    json.JSONDecodeError,
    aiohttp.ClientError,
    EnvoyHTTPStatusError,
)



"""Test endpoint for envoy v7 and newer firmware"""

import logging

import aiohttp
import pytest
from aioresponses import aioresponses
from syrupy.assertion import SnapshotAssertion

from pyenphase.models.dry_contacts import DryContactType

from .common import (
    get_mock_envoy,
    prep_envoy,
    start_7_firmware_mock,
)

LOGGER = logging.getLogger(__name__)


@pytest.mark.parametrize(
    (
        "version",
        "dry_contacts",
    ),
    [
        (
            "8.3.5167_3rd-pv",
            {
                "NC1": DryContactType.THIRD_PARTY_PV,
                "NC2": DryContactType.NONE,
                "NO1": DryContactType.NONE,
                "NO2": DryContactType.NONE,
            },
        ),
        (
            "8.2.127_with_generator_running",
            {
                "NC1": DryContactType.NONE,
                "NC2": DryContactType.NONE,
                "NO1": DryContactType.LOAD,
                "NO2": DryContactType.LOAD,
            },
        ),
    ],
    ids=[
        "8.3.5167_3rd-pv",
        "8.2.127_with_generator_running",
    ],
)
@pytest.mark.asyncio
async def test_dry_contact_type(
    version: str,
    snapshot: SnapshotAssertion,
    dry_contacts: dict[str, DryContactType],
    caplog: pytest.LogCaptureFixture,
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
) -> None:
    """Verify with 7.x firmware."""
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)
    caplog.set_level(logging.DEBUG)

    envoy = await get_mock_envoy(test_client_session)

    data = envoy.data
    assert data is not None

    for contact, type in dry_contacts.items():
        assert data.dry_contact_settings[contact].type == type



"""Model for an envoy."""

from dataclasses import dataclass, field
from typing import Any

from .acb import EnvoyACBPower, EnvoyBatteryAggregate
from .c6combiner import EnvoyC6CC
from .collar import EnvoyCollar
from .dry_contacts import EnvoyDryContactSettings, EnvoyDryContactStatus
from .encharge import EnvoyEncharge, EnvoyEnchargeAggregate, EnvoyEnchargePower
from .enpower import EnvoyEnpower
from .inverter import EnvoyInverter
from .meters import EnvoyMeterData
from .system_consumption import EnvoySystemConsumption
from .system_production import EnvoySystemProduction
from .tariff import EnvoyTariff


@dataclass(slots=True)
class EnvoyData:
    """
    Data Model for an envoy.

    Data is extract from raw data requested from Envoy. All raw
    data is also available as-received in :any:`raw`. For details
    on data models refer to the individual model descriptions.
    """

    #: dict of found Encharge device information, keyed by Encharge serial number.
    encharge_inventory: dict[str, EnvoyEncharge] | None = None
    #: dict of Enpower device power and charge state, keyed by Enpower serial number.
    encharge_power: dict[str, EnvoyEnchargePower] | None = None
    #: Aggregated data for all Encharge devices
    encharge_aggregate: EnvoyEnchargeAggregate | None = None
    #: EnchargePower device information
    enpower: EnvoyEnpower | None = None
    #: Power and soc for aggregated ACB batteries
    acb_power: EnvoyACBPower | None = None
    #: aggregated Enphase and ACB battery SOC and total capacity
    battery_aggregate: EnvoyBatteryAggregate | None = None
    #: IQ Meter collar, only for Envoy with IQ Meter Collar integrated consumption metering installed
    collar: EnvoyCollar | None = None
    #: Envoy C6 Combiner controller
    c6cc: EnvoyC6CC | None = None
    #: Consumption power & energy values, only for Envoy metered with CT installed
    system_consumption: EnvoySystemConsumption | None = None
    #: Solar Production power & energy values
    system_production: EnvoySystemProduction | None = None
    #: Individual phase consumption power & energy values, keyed by :any:`PhaseNames`,
    #: only for Envoy metered with CT installed
    system_consumption_phases: dict[str, EnvoySystemConsumption | None] | None = None
    #: Individual phase solar production power & energy values, keyed by :any:`PhaseNames`,
    #: only for Envoy metered with CT installed
    system_production_phases: dict[str, EnvoySystemProduction | None] | None = None
    #: Net consumption power & energy values, a.k.a. grid import/export,
    #: only for Envoy metered with CT installed
    system_net_consumption: EnvoySystemConsumption | None = None
    #: Individual phase Net consumption power & energy values, keyed by :any:`PhaseNames`,
    #: only for Envoy metered with CT installed
    system_net_consumption_phases: dict[str, EnvoySystemConsumption | None] | None = (
        None
    )
    #: CT power & energy values, only for Envoy metered with CT installed.
    #: Keyed by :any:`CtType`
    ctmeters: dict[str, EnvoyMeterData] = field(default_factory=dict)
    #: CT power & energy phase values, only for Envoy metered with CT installed.
    #: Keyed by :any:`CtType` and  :any:`PhaseNames`
    ctmeters_phases: dict[str, dict[str, EnvoyMeterData]] = field(default_factory=dict)
    # these are still here for backward compatibility
    #: Production CT power & energy values, only for Envoy metered with production CT installed
    #:
    #: May be deprecated in a future version, use :any:`ctmeters` [:any:`CtType.PRODUCTION`] instead
    #:
    ctmeter_production: EnvoyMeterData | None = None
    #: Consumption CT power & energy values, only for Envoy metered with consumption CT installed
    #:
    #: May be deprecated in a future version, use :any:`ctmeters` [:any:`CtType.TOTAL_CONSUMPTION`] or :any:`ctmeters` [:any:`CtType.NET_CONSUMPTION`] instead
    #:
    ctmeter_consumption: EnvoyMeterData | None = None
    #: Storage CT power & energy values, only for Envoy metered with storage CT installed
    #:
    #: May be deprecated in a future version, use :any:`ctmeters` [:any:`CtType.STORAGE`] instead
    #:
    ctmeter_storage: EnvoyMeterData | None = None
    #: Individual phase production CT power & energy values, keyed by :any:`PhaseNames`,
    #: only for Envoy metered with production CT installed
    #:
    #: May be deprecated in a future version, use :any:`ctmeters_phases` [:any:`CtType.PRODUCTION`] instead
    #:
    ctmeter_production_phases: dict[str, EnvoyMeterData] | None = None
    #: Individual phase consumption CT power & energy values, keyed by :any:`PhaseNames`,
    #: only for Envoy metered with consumption installed
    #:
    #: May be deprecated in a future version, use :any:`ctmeters_phases` [:any:`CtType.TOTAL_CONSUMPTION`] or :any:`ctmeters_phases` [:any:`CtType.NET_CONSUMPTION`] instead
    #:
    ctmeter_consumption_phases: dict[str, EnvoyMeterData] | None = None
    #: Individual phase storage CT power & energy values, keyed by :any:`PhaseNames`,
    #: only for Envoy metered with storage CT installed
    #:
    #: May be deprecated in a future version, use :any:`ctmeters_phases` [:any:`CtType.STORAGE`] instead
    #:
    ctmeter_storage_phases: dict[str, EnvoyMeterData] | None = None
    #: dict of Dry contact relay status, keyed by relay ID
    dry_contact_status: dict[str, EnvoyDryContactStatus] = field(default_factory=dict)
    #: dict of Dry contact relay settings, keyed by relay ID
    dry_contact_settings: dict[str, EnvoyDryContactSettings] = field(
        default_factory=dict
    )
    #: dict of Solar inverter data, keyed by inverter serial-number
    inverters: dict[str, EnvoyInverter] = field(default_factory=dict)
    #: Tariff information from Envoy
    tariff: EnvoyTariff | None = None
    # Raw data is exposed so we can __eq__ the data to see if
    # anything has changed and consumers of the library can
    # avoid dispatching data if nothing has changed.
    #: All request responses received from Envoy in last :any:`Envoy.update`, keyed by endpoint
    raw: dict[str, Any] = field(default_factory=dict)
